#include "la_5.h"
//#include <gl/glew.h>

/*
NUL4.0 - Nick's Best - www.nicksbest.com

@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@@  TNS OpenGL 2D/3D Rendering System  @@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@@                  __                 @@
@@            ___---  |                @@
@@      __----     @@  ||              @@
@@  _---  _         @@@  ||            @@
@@   |  @@@__         @@   ||          @@
@@    |  @@ #@@@#_     @@@   ||        @@
@@     |  @@     `#@@@@  @@    |_      @@
@@      |  @@         `#@@@@@_--       @@
@@       |  @@          _--'           @@
@@        |  @@     _---               @@
@@         |  @@_---                   @@
@@          |_--          _ __ __ __ __@@
@@                       |  Not Fancy  @@
@@                       |  But Easy   @@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

Author(s):WuYiming - xp8110@outlook.com

Want to join the development?
Append your name in the author list above.
Send feedback to la_support@nicksbest.com
*/

#include <math.h>
#include "freetype/ftadvanc.h"

char TNS_VERTEX_SIMPLE_MATCAP[] = "#version 330\n"
"uniform mat4 mProjection;\n"
"uniform mat4 mModel;\n"
"uniform mat4 mView;\n"
"\n"
"in vec4 vVertex;\n"
"in vec3 vNormal;\n"
"smooth out vec3 fNormal;\n"
"\n"
"void main(){\n"
"    gl_Position = mProjection * mView  * mModel * vVertex;\n"
"    vec3 N = ( mView * mModel * vec4(vNormal,0)).xyz;\n"
"    fNormal = normalize(N);\n"
"}";
char TNS_FRAGMENT_SIMPLE_MATCAP[] = "#version 330\n"
"\n"
"smooth in vec3 fNormal;"
"\n"
"float Interpolate(float between1,float between2,float value1,float value2,float key){\n"
"    float i = (key-between1)/(between2-between1);\n"
"    return value1*(1-i)+value2*i;"
"}\n"
"void main(){\n"
"    float value = dot(vec3(0,0,1),fNormal);\n"
"    if(value<0.65) value=0.15;\n"
"    else if(value>=0.65 && value<0.85) value=Interpolate(0.65,0.85,0.15,0.75,value);\n"
"    else if(value>=0.85 && value<0.95) value=0.75;\n"
"    else if(value>=0.95) value=0.9;\n"
"    gl_FragColor = vec4(vec3(0.84, 0.41, 0.16)*value,1);\n"
"}";

char TNS_VERTEX_GRID[] = "#version 330\n"
"\n"
"uniform mat4 mProjection;\n"
"uniform mat4 mModel;\n"
"uniform mat4 mView;\n"
"\n"
"in vec4 vVertex;\n"
"in vec4 vColor;\n"
"in vec2 vUV;\n"
"out vec4 fColor;\n"
"out vec2 uv;\n"
"\n"
"void main(){\n"
"    vec4 pos = mProjection * mView * mModel * vVertex;\n"
"    gl_Position = pos;\n"
"    fColor = vColor;\n"
"    uv = vUV;\n"
"}";

char TNS_FRAGMENT_TRANSPARNT_GRID[] = "#version 330\n"
"\n"
"in vec4 fColor;\n"
"in vec2 uv;\n"
"\n"
"void main(){\n"
"    vec4 c = fColor;\n"
"    c.a = sin(uv.x)*sin(uv.y)>0?c.a:0;\n"
"    gl_FragColor = c;\n"
"}";

const char LA_IMM_VERTEX_SHADER[] = "#version 330\n\
uniform mat4 mProjection;\
uniform mat4 mModel;\
uniform mat4 mView;\
in vec4 vVertex;\
in vec4 vColor;\
/*in vec4 vNormal;*/\
in vec2 vUV;\
out vec4 fColor;\
out vec2 fUV;\
void main(){\
    gl_Position = mProjection * mView * mModel * vVertex;\
    fColor = vColor;\
    fUV=vUV;\
}";
const char LA_IMM_FRAGMENT_SHADER[] = "#version 330\n\
uniform sampler2D TexColor;\
uniform sampler2DMS TexColorMS;\
uniform int TextureMode;\n\
uniform int MultiplyColor;\n\
uniform int SampleAmount;\n\
in vec4 fColor;\n\
in vec2 fUV;\n\
void main(){\n\
    vec4 color=vec4(1,0,1,1);\n\
    if(TextureMode==0){ color = fColor;}\n\
    else if(TextureMode==1){color = vec4(fColor.rgb,fColor.a*texture2D(TexColor,fUV.st).r);}\n\
    else if(TextureMode==2){\n\
        color=texture2D(TexColor,fUV.st);\n\
        if(MultiplyColor!=0){color*=fColor;}\n\
    }else if(TextureMode==3){\n\
        color=vec4(0,0,0,0);\n\
        ivec2 texSize = textureSize(TexColorMS);\n\
        for(int i=0;i<SampleAmount;i++) color+=texelFetch(TexColorMS, ivec2(fUV * texSize),i);\n\
        color/=SampleAmount;\n\
        if(MultiplyColor!=0){color*=fColor;}\n\
    }\n\
    gl_FragColor = color;\n\
}";
const char LA_RAY_VERTEX_SHADER[] = "#version 330\n\
in vec3 vUV;\n\
in vec4 vVertex;\n\
out vec3 fViewDir;\n\
void main(){\n\
    gl_Position=vVertex;\n\
    fViewDir = vUV;\n\
}";
const char LA_RAY_FRAGMENT_SHADER[] = "#version 330\n\
uniform vec3 uViewDir;\n\
uniform vec3 uViewPos;\n\
uniform float uFOV;\n\
in vec3 fViewDir;\n\
void main(){\n\
    float d=dot(uViewDir,normalize(fViewDir));\n\
    float target=cos(uFOV/2.);\n\
    vec4 color=vec4(1.,1.,1.,1.); float mul=0.;\n\
    if(d<(target+0.005)&&d>target) mul=1.0;\n\
    gl_FragColor = mul*color;\n\
}";
const char LA_SCENE_VERTEX_SHADER[] = "#version 330\n\
uniform mat4 mProjection;\n\
uniform mat4 mModel;\n\
uniform mat4 mView;\n\
uniform mat4 mShadow;\n\
in vec4 vVertex;\n\
in vec4 vColor;\n\
in vec4 vNormal;\n\
in vec2 vUV;\n\
out vec4 fColor;\n\
//out vec4 fNormal;\n\
out vec2 fUV;\n\
out vec4 fGPos;\n\
void main(){\n\
    gl_Position= mProjection * mView * mModel * vVertex;\n\
    fUV=vUV;\n\
    //fNormal=vNormal;\n\
    fColor=vColor;\n\
    fGPos= mShadow * mModel * vVertex;\
}";
const char LA_SCENE_FRAGMENT_SHADER[] = "#version 330\n\
uniform sampler2D TexColor;\n\
uniform sampler2DMS TexColorMS;\
uniform int TextureMode;\n\
uniform int SampleAmount;\n\
uniform int MultiplyColor;\n\
in vec4 fColor;\n\
//in vec4 fNormal;\n\
in vec2 fUV;\n\
in vec4 fGPos;\n\
vec4 GetTexture(vec2 uv){\n\
    vec4 color=vec4(1,0,1,1);\n\
    if(TextureMode==1 || TextureMode==2){ return texture2D(TexColor,uv); }\n\
    else if(TextureMode==3){\n\
        ivec2 texSize = textureSize(TexColorMS);\n\
        for(int i=0;i<SampleAmount;i++) color+=texelFetch(TexColorMS, ivec2(fUV * texSize),i);\n\
        color/=SampleAmount;\n\
        if(MultiplyColor!=0){color*=fColor;}\n\
        return color;\n\
    }\n\
    else return vec4(1,0,1,1);\n\
}\n\
float GetShadow(vec4 GPos){\n\
    vec3 projCoords = GPos.xyz / GPos.w;\n\
    projCoords = projCoords * 0.5 + 0.5;\n\
    float closestDepth = GetTexture(projCoords.xy).r;\n\
    float currentDepth = projCoords.z;\n\
    float shadow = currentDepth > (closestDepth+0.001)  ? 0.5 : 1.0;\n\
    return shadow;\n\
}\n\
void main(){\n\
    gl_FragColor=GetShadow(fGPos)*fColor;\n\
}";

const char LA_CASCADE_SHADOW_VERTEX_SHADER[] = "#version 330\n\
in vec4 vVertex;\n\
uniform mat4 mModel;\n\
uniform mat4 mShadow;\n\
void main(){\n\
    gl_Position=mShadow*mModel*vVertex;\n\
}";
const char LA_CASCADE_SHADOW_FRAGMENT_SHADER[] = "#version 330\nvoid main(){gl_FragDepth = gl_FragCoord.z;}";

const char LA_SELECTION_VERTEX_SHADER[] = "#version 330\n\
in vec4 vVertex;\n\
in vec3 vColor;\n\
uniform mat4 mProjection;\n\
uniform mat4 mModel;\n\
uniform mat4 mView;\n\
flat out vec3 fIdColor;\n\
void main(){\n\
    gl_Position = mProjection * mView * mModel * vVertex;\n\
    fIdColor = vColor;\n\
}";
const char LA_SELECTION_FRAGMENT_SHADER[] = "#version 330\n\
flat in vec3 fIdColor;\n\
void main(){\n\
    gl_FragColor=vec4(fIdColor,1.);\n\
}";

int tKnlAttatchShader(tnsShader *tns);
tnsShader *tKnlFindShader1i(int CustomIndex);
void tKnlPushMatrix();
void tKnlPopMatrix();

laListHandle *tKnlGetTextureList();

void tnsLoadIdentity44d(tnsMatrix44d m);
void tnsMakeOrthoMatrix44d(tnsMatrix44d mProjection, real xMin, real xMax, real yMin, real yMax, real zMin, real zMax);
void tnsMakePerspectiveMatrix44d(tnsMatrix44d mProjection, real fFov_rad, real fAspect, real zMin, real zMax);
void tnsMakeTranslationMatrix44d(tnsMatrix44d mTrans, real x, real y, real z);
void tnsMakeRotationMatrix44d(tnsMatrix44d m, real angle_rad, real x, real y, real z);
void tnsMakeScaleMatrix44d(tnsMatrix44d m, real x, real y, real z);
void tnsMultiply44d(tnsMatrix44d result, tnsMatrix44d l, tnsMatrix44d r);

void tnsInitFirstLevel(tnsMatrixStack *tms);
tnsMatrixStackItem *tKnlGetCurrentMatStackItem();
tnsShader *tKnlGetActiveShader();

void tnsAttach2DOffscreenBuffer(tnsOffscreen *target, GLuint attatchment, tnsTexture *use);
void tnsDetach2DOffscreenBuffer(tnsOffscreen *target, GLuint which_attach_point);

//=========================================================================================

tnsMain *T;
real DefaultZRange[2] = {0.1, 1000};

//=======================================[SYS]
extern LA MAIN;

void InitGLRenderEnviornment(){
    glEnable(GL_SCISSOR_TEST);
};
void tnsSwitchToCurrentWindowContext(void *wnd){
    //HGLRC current = wglGetCurrentContext(), hglrc = wnd->SystemGLRC;
    //HDC hdc = wnd->SystemDC;
//
    ////if (hglrc != current)
    //int a = GetLastError();
//
    //int s = wglMakeCurrent(hdc, hglrc);
    laWindow* win = wnd;
    glXMakeContextCurrent(MAIN.dpy, win->win, win->win, MAIN.glc);
};
void tnsViewportWithScissor(int x, int y, int w, int h){
    tnsShader *current_shader = 0;
    glEnable(GL_SCISSOR_TEST);
    glViewport(x, y, w, h);
    glScissor(x, y, w, h);
    if ((current_shader = tKnlGetActiveShader())){
        tnsResetViewMatrix();
        tnsShaderApplyView(current_shader, tnsGetViewMatrix());
        tnsResetModelMatrix();
        tnsShaderApplyModel(current_shader, tnsGetModelMatrix());
    }
}
void tnsViewport(int x, int y, int w, int h){
    glViewport(x, y, w, h);
}
//Must At Origion!

//void UseWindowMatrix_WithScissor(laWndPlacement* wp){
//	tnsViewportWithScissor(0, 0, VAL2_CLIENT_W_H(*wp));
//	SetMatrix2Di(0, VAL2_CLIENT_W_H(*wp), 0);
//};
//void UseBlockMatrix_WithScissor(laBlockPlacement* bp){
//	tnsViewportWithScissor(bp->UpperLeftX,
//		bp->ClientHeight - bp->LowerRightY,
//		bp->LowerRightX - bp->UpperLeftX,
//		bp->LowerRightY - bp->UpperLeftY);
//	SetMatrix2Di(0, bp->LowerRightX - bp->UpperLeftX,
//		0, bp->UpperLeftY - bp->LowerRightY);
//};

void DrawWireRect2dp(real x, real y, real x2, real y2){
    real Verts[8];
    tnsMakeQuad2d(Verts, x, y, x2, y, x2, y2, x, y2);
    tnsVertexArray2d(Verts, 4);
    tnsPackAs(GL_LINE_LOOP);
};
void DrawWireRect4ds(real x, real y, real w, real h){
    real Verts[8];
    tnsMakeQuad2d(Verts, x, y, x + w, y, x + w, y + h, x, y + h);
    tnsVertexArray2d(Verts, 4);
    tnsPackAs(GL_LINE_LOOP);
};
void DrawWireRect4drs(real x, real y, real w, real h){
    real Verts[8];
    tnsMakeQuad2d(Verts, x, y, x + w, y, x + w, y - h, x, y - h);
    tnsVertexArray2d(Verts, 4);
    tnsPackAs(GL_LINE_LOOP);
};
void DrawSoildRect4drs(real x, real y, real w, real h){
    real Verts[8];
    tnsMakeQuad2d(Verts, x, y, x + w, y, x + w, y - h, x, y - h);
    tnsVertexArray2d(Verts, 4);
    tnsPackAs(GL_TRIANGLE_FAN);
}
void DrawWireCross4ds(real x, real y, real w, real h){
    real Verts[8];
    tnsMakeQuad2d(Verts, x, y, x + w, y + h, x + w, y, x, y + h);
    tnsVertexArray2d(Verts, 4);
    tnsPackAs(GL_LINES);
};
void DrawWireCross4drs(real x, real y, real w, real h){
    real Verts[8];
    tnsMakeQuad2d(Verts, x, y, x + w, y - h, x + w, y, x, y - h);
    tnsVertexArray2d(Verts, 4);
    tnsPackAs(GL_LINES);
};
void DrawGrid6f(real L, real R, real U, real B, real stepX, real stepY){
    //do nothing;
}
void DrawSolidArrow1i3d(int direction, real xoff, real yoff, real size){
    real size_2 = size / 2.0f;
    real Verts[6];

    switch (direction){
    case TNS_ARROW_L:
        tnsMakeTriangle(Verts,
                        xoff + size_2, yoff,
                        xoff, yoff - size_2,
                        xoff + size_2, yoff - size);
        break;
    case TNS_ARROW_R:
        tnsMakeTriangle(Verts,
                        xoff, yoff,
                        xoff + size_2, yoff - size_2,
                        xoff, yoff - size);
        break;
    }

    tnsPackAs(GL_TRIANGLES);
}

//=======================================[Shader]

int tnsNextPowOf2(int i){
    int result = 2;
    while (result < i){
        result *= 2;
    }
    return result;
}

void tnsShaderMakeIndex(tnsShader *tns){
    int program;

    if (!tns) return;

    program = tns->glProgramID;

    tns->iModel = glGetUniformLocation(program, "mModel");
    tns->iProjection = glGetUniformLocation(program, "mProjection");
    tns->iProjectionInverse = glGetUniformLocation(program, "mProjectionInverse");
    tns->iView = glGetUniformLocation(program, "mView");
    tns->iNormal = glGetUniformLocation(program, "mNormalScaler");
    tns->iShadow = glGetUniformLocation(program, "mShadow");

    tns->iVertex = glGetAttribLocation(program, "vVertex");
    tns->iNormal = glGetAttribLocation(program, "vNormal");
    tns->iColor = glGetAttribLocation(program, "vColor");
    tns->iUV = glGetAttribLocation(program, "vUV");

    tns->iTexColor = glGetUniformLocation(program, "TexColor");
    tns->iTexColorMS = glGetUniformLocation(program, "TexColorMS");
    tns->iMultiplyColor = glGetUniformLocation(program, "MultiplyColor");
    tns->iTextureMode = glGetUniformLocation(program, "TextureMode");
    tns->iSampleAmount = glGetUniformLocation(program, "SampleAmount");
    if(tns->iTexColor>=0){glUniform1i(tns->iTexColor, 0);}
    if(tns->iTexColorMS>=0){glUniform1i(tns->iTexColorMS, 1);}

    tns->uViewDir = glGetUniformLocation(program, "uViewDir");
    tns->uViewPos = glGetUniformLocation(program, "uViewPos");
    tns->uFOV = glGetUniformLocation(program, "uFOV");
}
void tnsShaderApplyProjection(tnsShader *tns, tnsMatrix44d m){
    tnsMatrix44f mf;
    if (T->BindedShader != tns){ glUseProgram(tns->glProgramID); T->BindedShader = tns; }
    tnsConvert44df(m, mf);
    glUniformMatrix4fv(tns->iProjection, 1, 0, mf);
}
void tnsShaderApplyProjectionInverse(tnsShader *tns, tnsMatrix44d m){
    tnsMatrix44f mf;
    tnsMatrix44d i;
    if (T->BindedShader != tns){ glUseProgram(tns->glProgramID); T->BindedShader = tns; }
    tnsInverse44d(i, m);
    tnsConvert44df(i, mf);
    glUniformMatrix4fv(tns->iProjectionInverse, 1, 0, mf);
}
void tnsShaderApplyModel(tnsShader *tns, tnsMatrix44d m){
    tnsMatrix44f mf;
    if (T->BindedShader != tns){ glUseProgram(tns->glProgramID); T->BindedShader = tns; }
    tnsConvert44df(m, mf);
    glUniformMatrix4fv(tns->iModel, 1, 0, mf);
}
void tnsShaderApplyView(tnsShader *tns, tnsMatrix44d m){
    tnsMatrix44f mf;
    if (T->BindedShader != tns){ glUseProgram(tns->glProgramID); T->BindedShader = tns; }
    tnsConvert44df(m, mf);
    glUniformMatrix4fv(tns->iView, 1, 0, mf);
}
void tnsShaderApplyNormalScaler(tnsShader *tns, tnsMatrix44d m){
    tnsMatrix44f mf;
    if (tns->iNormal == -1) return;
    if (T->BindedShader != tns){ glUseProgram(tns->glProgramID); T->BindedShader = tns; }
    tnsConvert44df(m, mf);
    glUniformMatrix4fv(tns->iNormal, 1, 0, mf);
}
void tnsShaderApplyShadowMatrix(tnsShader *tns, tnsMatrix44d m){
    tnsMatrix44f mf;
    if (tns->iShadow == -1) return;
    if (T->BindedShader != tns){ glUseProgram(tns->glProgramID); T->BindedShader = tns; }
    tnsConvert44df(m, mf);
    glUniformMatrix4fv(tns->iShadow, 1, 0, mf);
}

//--------------------Export
int tnsNewVertexShader(char *Content){
    int status = 0;
    char error[1024];
    GLuint VertexShaderObject;
    tnsShader *s = 0;

    if (!Content) return -1;

    VertexShaderObject = glCreateShader(GL_VERTEX_SHADER);

    glShaderSource(VertexShaderObject, 1, &Content, 0);
    glCompileShader(VertexShaderObject);
    glGetShaderiv(VertexShaderObject, GL_COMPILE_STATUS, &status);
    if (status == GL_FALSE){
        glGetShaderInfoLog(VertexShaderObject, sizeof(error), 0, error);
        printf("Vertex shader error:\n%s", error);
        glDeleteShader(VertexShaderObject);
        return -1;
    }

    return VertexShaderObject;
}
int tnsNewFragmentShader(char *Content){
    int status = 0;
    char error[1024];
    GLuint FragmentShaderObject;
    tnsShader *s = 0;

    if (!Content) return -1;

    FragmentShaderObject = glCreateShader(GL_FRAGMENT_SHADER);

    glShaderSource(FragmentShaderObject, 1, &Content, 0);
    glCompileShader(FragmentShaderObject);
    glGetShaderiv(FragmentShaderObject, GL_COMPILE_STATUS, &status);
    if (status == GL_FALSE){
        glGetShaderInfoLog(FragmentShaderObject, sizeof(error), 0, error);
        printf("Fragment shader error:\n%s", error);
        glDeleteShader(FragmentShaderObject);
        return -1;
    }

    return FragmentShaderObject;
}
int tnsNewGeometryShader(char *Content){
    int status = 0;
    char error[1024];
    GLuint GeometryShaderObject;
    tnsShader *s = 0;

    if (!Content) return -1;

    GeometryShaderObject = glCreateShader(GL_GEOMETRY_SHADER);

    glShaderSource(GeometryShaderObject, 1, &Content, 0);
    glCompileShader(GeometryShaderObject);
    glGetShaderiv(GeometryShaderObject, GL_COMPILE_STATUS, &status);
    if (status == GL_FALSE){
        glGetShaderInfoLog(GeometryShaderObject, sizeof(error), 0, error);
        printf("Geometry shader error:\n%s", error);
        glDeleteShader(GeometryShaderObject);
        return -1;
    }

    return GeometryShaderObject;
}

tnsShader *tnsNewShaderProgram(int VertexShaderID, int FragmentShaderID, int GeometryShaderID){
    int vso = VertexShaderID;
    int fso = FragmentShaderID;
    int gso = GeometryShaderID;
    tnsShader *tns = 0;
    int status = 0;
    char error[1024];

    if (!vso || !fso) return 0;

    tns = CreateNew(tnsShader);
    tns->vtShaderID = vso;
    tns->fgShaderID = fso;
    tns->glProgramID = glCreateProgram();
    glAttachShader(tns->glProgramID, vso);
    glAttachShader(tns->glProgramID, fso);
    if (GeometryShaderID > -1){ glAttachShader(tns->glProgramID, gso); tns->gsShaderID=gso; }
    glLinkProgram(tns->glProgramID);
    glGetProgramiv(tns->glProgramID, GL_LINK_STATUS, &status);
    if (status == GL_FALSE){
        glGetProgramInfoLog(tns->glProgramID, sizeof(error), 0, error);
        printf("Shader Linking error:\n%s", error);
        glDetachShader(tns->glProgramID, vso);
        glDetachShader(tns->glProgramID, fso);
        glDeleteShader(vso);
        glDeleteShader(fso);
        glDeleteProgram(tns->glProgramID);
        FreeMem(tns);
        return 0;
    }
    glUseProgram(tns->glProgramID);

    tnsShaderMakeIndex(tns);

    return tns;
}
int tnsEnableShader(int index){
    tnsMatrixStackItem *tmsi;
    tnsShader *tns = tKnlFindShader1i(index);
    if (!tns){
        glUseProgram(0);
        T->CurrentShader = 0;
        T->BindedShader = 0;
        return 0;
    }
    glUseProgram(tns->glProgramID);
    T->CurrentShader = tns;
    T->BindedShader = tns;

    tmsi = tKnlGetCurrentMatStackItem();
    tnsShaderApplyProjection(tns, tmsi->projection);
    tnsShaderApplyView(tns, tmsi->view);
    tnsShaderApplyModel(tns, tmsi->model);
    return 1;
}
int tnsEnableShaderv(tnsShader *shader){
    tnsMatrixStackItem *tmsi;
    tnsShader *tns = shader;
    if (!tns){
        glUseProgram(0);
        T->CurrentShader = 0;
        T->BindedShader = 0;
        return 0;
    }
    glUseProgram(tns->glProgramID);
    T->CurrentShader = tns;
    T->BindedShader = tns;

    tmsi = tKnlGetCurrentMatStackItem();
    tnsShaderApplyProjection(tns, tmsi->projection);
    tnsShaderApplyProjectionInverse(tns, tmsi->projection);
    tnsShaderApplyView(tns, tmsi->view);
    tnsShaderApplyModel(tns, tmsi->model);

    //if (tns->iVertex != -1) glEnableVertexAttribArray(tns->iVertex);
    //if (tns->iColor != -1) glEnableVertexAttribArray(tns->iColor);
    //if (tns->iNormal != -1) glEnableVertexAttribArray(tns->iNormal);
    //if (tns->iUV != -1) glEnableVertexAttribArray(tns->iUV);

    return 1;
}
int tnsUseShader(tnsShader *shader){
    T->StateShader = shader;
}
void tnsDeleteShaderProgram(tnsShader* s){
    tnsUseShader(0); tnsEnableShaderv(0);
    if(s->vtShaderID>-1) glDeleteShader(s->vtShaderID);
    if(s->fgShaderID>-1) glDeleteShader(s->fgShaderID);
    if(s->gsShaderID>-1) glDeleteShader(s->gsShaderID);
    glDeleteProgram(s->glProgramID);
    free(s);
}

void tnsUseImmShader(){
    T->StateShader = T->immShader;
}
void tnsUseTransparentGridShader(){
    T->StateShader = T->TransparentGridShader;
}
void tnsUseExtraBufferShader(){
    T->StateShader = T->ExtraBuffersShader;
}
void tnsUseRayShader(){
    T->StateShader = T->RayShader;
}
void tnsUseShadowShader(){
    T->StateShader = T->ShadowShader;
}
void tnsUseSceneShader(){
    T->StateShader = T->SceneShader;
}

//=======================================[MAT]

int tnsLineIntersectTest2d(tnsVector2d a1, tnsVector2d a2, tnsVector2d b1, tnsVector2d b2, double *aRatio){
    double k1, k2;
    double x;
    double y;
    double Ratio;
    double xDiff = (a2[0] - a1[0]); // +DBL_EPSILON;
    double xDiff2 = (b2[0] - b1[0]);

    if (xDiff == 0){
        if (xDiff2 == 0){
            *aRatio = 0;
            return 0;
        }
        double r2 = tnsGetRatiod(b1[0], b2[0], a1[0]);
        y = tnsLinearItp(b1[1], b2[1], r2);
        *aRatio = Ratio = tnsGetRatiod(a1[1], a2[1], y);
    }else{
        if (xDiff2 == 0){
            Ratio = tnsGetRatiod(a1[0], a2[0], b1[0]);
            //y = tnsLinearItp(a1[1], a2[1], r2);
            *aRatio = Ratio;
        }else{
            k1 = (a2[1] - a1[1]) / xDiff;
            k2 = (b2[1] - b1[1]) / xDiff2;

            if ((k1 == k2)) return 0;

            x = (a1[1] - b1[1] - k1 * a1[0] + k2 * b1[0]) / (k2 - k1);

            Ratio = (x - a1[0]) / xDiff;

            *aRatio = Ratio;
        }
    }

    if (b1[0] == b2[0]){
        y = tnsLinearItp(a1[1], a2[1], Ratio);
        if (y > TNS_MAX2(b1[1], b2[1]) || y < TNS_MIN2(b1[1], b2[1])) return 0;
    }else if (Ratio <= 0 || Ratio > 1 ||
             (b1[0] > b2[0] && x > b1[0]) ||
             (b1[0] < b2[0] && x < b1[0]) ||
             (b2[0] > b1[0] && x > b2[0]) ||
             (b2[0] < b1[0] && x < b2[0]))
        return 0;

    return 1;
}
double tnsGetLineZ(tnsVector3d L, tnsVector3d R, real Ratio){
    //double z = 1 / tnsLinearItp(1 / L[2], 1 / R[2], Ratio);
    double z = tnsLinearItp(L[2], R[2], Ratio);
    return z;
}
double tnsGetLineZPoint(tnsVector3d L, tnsVector3d R, tnsVector3d FromL){
    double r = (FromL[0] - L[0]) / (R[0] - L[0]);
    return tnsLinearItp(L[2], R[2], r);
    //return 1 / tnsLinearItp(1 / L[2], 1 / R[2], r);
}
double tnsGetRatio3d(tnsVector3d L, tnsVector3d R, tnsVector3d FromL){
    double r = (FromL[0] - L[0]) / (R[0] - L[0]);
    return r;
}

double tnsGetRatiod(real L, real R, real FromL){
    double r = (FromL - L) / (R - L);
    return r;
}

void tnsInterpolateTripple2d(tnsVector2d v1, tnsVector2d v2, tnsVector2d v3, real ratio, tnsVector2d result){
    tnsVector2d i1,i2;
    tnsInterpolate2dv(v1,v2,ratio,i1);
    tnsInterpolate2dv(v2,v3,ratio,i2);
    tnsInterpolate2dv(i1,i2,ratio,result);
}

void tnsVectorMinus2d(tnsVector2d result, tnsVector2d l, tnsVector2d r){
    result[0] = l[0] - r[0];
    result[1] = l[1] - r[1];
}
void tnsVectorMinus3d(tnsVector3d result, tnsVector3d l, tnsVector3d r){
    result[0] = l[0] - r[0];
    result[1] = l[1] - r[1];
    result[2] = l[2] - r[2];
}
void tnsVectorSubtract3d(tnsVector3d l, tnsVector3d r){
    l[0] = l[0] - r[0];
    l[1] = l[1] - r[1];
    l[2] = l[2] - r[2];
}
void tnsVectorPlus3d(tnsVector3d result, tnsVector3d l, tnsVector3d r){
    result[0] = l[0] + r[0];
    result[1] = l[1] + r[1];
    result[2] = l[2] + r[2];
}
void tnsVectorAccum3d(tnsVector3d l, tnsVector3d r){
    l[0] = l[0] + r[0];
    l[1] = l[1] + r[1];
    l[2] = l[2] + r[2];
}
void tnsVectorAccum2d(tnsVector2d l, tnsVector2d r){
    l[0] = l[0] + r[0];
    l[1] = l[1] + r[1];
}
void tnsVectorNegate3d(tnsVector3d result, tnsVector3d l){
    result[0] = -l[0];
    result[1] = -l[1];
    result[2] = -l[2];
}
void tnsVectorNegateSelf3d(tnsVector3d l){
    l[0] = -l[0];
    l[1] = -l[1];
    l[2] = -l[2];
}
void tnsVectorCopy2d(tnsVector2d from, tnsVector2d to){
    to[0] = from[0];
    to[1] = from[1];
}
void tnsVectorCopy3d(tnsVector3d from, tnsVector3d to){
    to[0] = from[0];
    to[1] = from[1];
    to[2] = from[2];
}
void tnsVectorCopy4d(tnsVector4d from, tnsVector4d to){
    to[0] = from[0];
    to[1] = from[1];
    to[2] = from[2];
    to[3] = from[3];
}
void tnsVectorMultiSelf4d(tnsVector4d from, real num){
    from[0] *= num;
    from[1] *= num;
    from[2] *= num;
    from[3] *= num;
}
void tnsVectorMultiSelf3d(tnsVector3d from, real num){
    from[0] *= num;
    from[1] *= num;
    from[2] *= num;
}
void tnsVectorMultiSelf2d(tnsVector2d from, real num){
    from[0] *= num;
    from[1] *= num;
}
void tnsVectorMulti4d(tnsVector4d to, tnsVector4d from, real num){
    to[0]=from[0]*num;
    to[1]=from[1]*num;
    to[2]=from[2]*num;
    to[3]=from[3]*num;
}
void tnsVectorMulti3d(tnsVector3d to, tnsVector3d from, real num){
    to[0]=from[0]*num;
    to[1]=from[1]*num;
    to[2]=from[2]*num;
}
void tnsVectorMulti2d(tnsVector2d to, tnsVector2d from, real num){
    to[0]=from[0]*num;
    to[1]=from[1]*num;
}

real tnsDirectionToRad(tnsVector2d Dir){
    real arcc = acos(Dir[0]);
    real arcs = asin(Dir[1]);

    if (Dir[0] >= 0){
        if (Dir[1] >= 0) return arcc;
        else
            return TNS_PI * 2 - arcc;
    }else{
        if (Dir[1] >= 0) return arcs + TNS_PI / 2;
        else
            return TNS_PI + arcs;
    }
}
void tnsConvert44df(tnsMatrix44d from, tnsMatrix44f to){
    to[0] = from[0];
    to[1] = from[1];
    to[2] = from[2];
    to[3] = from[3];
    to[4] = from[4];
    to[5] = from[5];
    to[6] = from[6];
    to[7] = from[7];
    to[8] = from[8];
    to[9] = from[9];
    to[10] = from[10];
    to[11] = from[11];
    to[12] = from[12];
    to[13] = from[13];
    to[14] = from[14];
    to[15] = from[15];
}

void tnsLoadIdentity44d(tnsMatrix44d m){
    memset(m, 0, sizeof(tnsMatrix44d));
    m[0] = 1.0f;
    m[5] = 1.0f;
    m[10] = 1.0f;
    m[15] = 1.0f;
};

real tnsDistIdv2(real x1, real y1, real x2, real y2){
    real x = x2 - x1, y = y2 - y1;
    return sqrt((x * x + y * y));
}
real tnsDist3dv(tnsVector3d l, tnsVector3d r){
    real x = r[0] - l[0];
    real y = r[1] - l[1];
    real z = r[2] - l[2];
    return sqrt(x * x + y * y + z * z);
}
real tnsDist2dv(tnsVector2d l, tnsVector2d r){
    real x = r[0] - l[0];
    real y = r[1] - l[1];
    return sqrt(x * x + y * y);
}

real tnsLength3d(tnsVector3d l){
    return (sqrt(l[0] * l[0] + l[1] * l[1] + l[2] * l[2]));
}
real tnsLength2d(tnsVector3d l){
    return (sqrt(l[0] * l[0] + l[1] * l[1]));
}
void tnsNormalize3d(tnsVector3d result, tnsVector3d l){
    real r = sqrt(l[0] * l[0] + l[1] * l[1] + l[2] * l[2]);
    result[0] = l[0] / r;
    result[1] = l[1] / r;
    result[2] = l[2] / r;
}
void tnsNormalize2d(tnsVector2d result, tnsVector2d l){
    real r = sqrt(l[0] * l[0] + l[1] * l[1]);
    result[0] = l[0] / r;
    result[1] = l[1] / r;
}
void tnsNormalizeSelf2d(tnsVector3d result){
    real r = sqrt(result[0] * result[0] + result[1] * result[1]);
    result[0] /= r;
    result[1] /= r;
}
void tnsNormalizeSelf3d(tnsVector3d result){
    real r = sqrt(result[0] * result[0] + result[1] * result[1] + result[2] * result[2]);
    result[0] /= r;
    result[1] /= r;
    result[2] /= r;
}
real tnsDot3d(tnsVector3d l, tnsVector3d r, int normalize){
    tnsVector3d ln, rn;
    if (normalize){
        tnsNormalize3d(ln, l);
        tnsNormalize3d(rn, r);
        return (ln[0] * rn[0] + ln[1] * rn[1] + ln[2] * rn[2]);
    }
    return (l[0] * r[0] + l[1] * r[1] + l[2] * r[2]);
}
real tnsDot2d(tnsVector2d l, tnsVector2d r, int normalize){
    tnsVector3d ln, rn;
    if (normalize){
        tnsNormalize2d(ln, l);
        tnsNormalize2d(rn, r);
        return (ln[0] * rn[0] + ln[1] * rn[1]);
    }
    return (l[0] * r[0] + l[1] * r[1]);
}
real tnsVectorCross3d(tnsVector3d result, tnsVector3d l, tnsVector3d r){
    result[0] = l[1] * r[2] - l[2] * r[1];
    result[1] = l[2] * r[0] - l[0] * r[2];
    result[2] = l[0] * r[1] - l[1] * r[0];
    return tnsLength3d(result);
}
void tnsVectorCrossOnly3d(tnsVector3d result, tnsVector3d l, tnsVector3d r){
    result[0] = l[1] * r[2] - l[2] * r[1];
    result[1] = l[2] * r[0] - l[0] * r[2];
    result[2] = l[0] * r[1] - l[1] * r[0];
}
real tnsAngleRad3d(tnsVector3d from, tnsVector3d to, tnsVector3d PositiveReference){
    if (PositiveReference){
        tnsVector3d res;
        tnsVectorCross3d(res, from, to);
        if (tnsDot3d(res, PositiveReference, 1) > 0) return acosf(tnsDot3d(from, to, 1));
        else
            return -acosf(tnsDot3d(from, to, 1));
    }
    return acosf(tnsDot3d(from, to, 1));
}
void tnsApplyRotation33d(tnsVector3d result, tnsMatrix44d mat, tnsVector3d v){
    result[0] = mat[0] * v[0] + mat[1] * v[1] + mat[2] * v[2];
    result[1] = mat[3] * v[0] + mat[4] * v[1] + mat[5] * v[2];
    result[2] = mat[6] * v[0] + mat[7] * v[1] + mat[8] * v[2];
}
void tnsApplyRotation43d(tnsVector3d result, tnsMatrix44d mat, tnsVector3d v){
    result[0] = mat[0] * v[0] + mat[1] * v[1] + mat[2] * v[2];
    result[1] = mat[4] * v[0] + mat[5] * v[1] + mat[6] * v[2];
    result[2] = mat[8] * v[0] + mat[9] * v[1] + mat[10] * v[2];
}
void tnsApplyTransform43d(tnsVector3d result, tnsMatrix44d mat, tnsVector3d v){
    real w;
    result[0] = mat[0] * v[0] + mat[4] * v[1] + mat[8] * v[2] + mat[12] * 1;
    result[1] = mat[1] * v[0] + mat[5] * v[1] + mat[9] * v[2] + mat[13] * 1;
    result[2] = mat[2] * v[0] + mat[6] * v[1] + mat[10] * v[2] + mat[14] * 1;
    w = mat[3] * v[0] + mat[7] * v[1] + mat[11] * v[2] + mat[15] * 1;
    //result[0] /= w;
    //result[1] /= w;
    //result[2] /= w;
}
void tnsApplyNormalTransform43d(tnsVector3d result, tnsMatrix44d mat, tnsVector3d v){
    real w;
    result[0] = mat[0] * v[0] + mat[4] * v[1] + mat[8] * v[2] + mat[12] * 1;
    result[1] = mat[1] * v[0] + mat[5] * v[1] + mat[9] * v[2] + mat[13] * 1;
    result[2] = mat[2] * v[0] + mat[6] * v[1] + mat[10] * v[2] + mat[14] * 1;
}
void tnsApplyTransform44d(tnsVector4d result, tnsMatrix44d mat, tnsVector4d v){
    result[0] = mat[0] * v[0] + mat[4] * v[1] + mat[8] * v[2] + mat[12] * 1;
    result[1] = mat[1] * v[0] + mat[5] * v[1] + mat[9] * v[2] + mat[13] * 1;
    result[2] = mat[2] * v[0] + mat[6] * v[1] + mat[10] * v[2] + mat[14] * 1;
    result[3] = mat[3] * v[0] + mat[7] * v[1] + mat[11] * v[2] + mat[15] * 1;
}
void tnsApplyTransform44dTrue(tnsVector4d result, tnsMatrix44d mat, tnsVector4d v){
    result[0] = mat[0] * v[0] + mat[4] * v[1] + mat[8] * v[2] + mat[12] * v[3];
    result[1] = mat[1] * v[0] + mat[5] * v[1] + mat[9] * v[2] + mat[13] * v[3];
    result[2] = mat[2] * v[0] + mat[6] * v[1] + mat[10] * v[2] + mat[14] * v[3];
    result[3] = mat[3] * v[0] + mat[7] * v[1] + mat[11] * v[2] + mat[15] * v[3];
}

void tnsRemoveTranslation44d(tnsMatrix44d result, tnsMatrix44d mat){
    tnsLoadIdentity44d(result);
    result[0] = mat[0];
    result[1] = mat[1];
    result[2] = mat[2];
    result[4] = mat[4];
    result[5] = mat[5];
    result[6] = mat[6];
    result[8] = mat[8];
    result[9] = mat[9];
    result[10] = mat[10];
}
void tnsClearTranslation44d(tnsMatrix44d mat){
    mat[3] = 0;
    mat[7] = 0;
    mat[11] = 0;
}

void tnsExtractXYZEuler44d(tnsMatrix44d mat, real *x_result, real *y_result, real *z_result){
    real xRot, yRot, zRot;

    if (mat[2] < 1){
        if (mat[2] > -1){
            yRot = asin(mat[2]);
            xRot = atan2(-mat[6], mat[10]);
            zRot = atan2(-mat[1], mat[0]);
        }else{
            yRot = -TNS_PI / 2;
            xRot = -atan2(-mat[4], mat[5]);
            zRot = 0;
        }
    }else{
        yRot = TNS_PI / 2;
        xRot = atan2(-mat[4], mat[5]);
        zRot = 0;
    }

    (*x_result) = -xRot;
    (*y_result) = -yRot;
    (*z_result) = -zRot;
}
void tnsExtractLocation44d(tnsMatrix44d mat, real *x_result, real *y_result, real *z_result){
    *x_result = mat[12];
    *y_result = mat[13];
    *z_result = mat[14];
}
void tnsExtractUniformScale44d(tnsMatrix44d mat, real *result){
    tnsVector3d v = {mat[0], mat[1], mat[2]};
    *result = tnsLength3d(v);
}

#define L(row, col) l[(col << 2) + row]
#define R(row, col) r[(col << 2) + row]
#define P(row, col) result[(col << 2) + row]

void tnsPrintMatrix44d(tnsMatrix44d l){
    int i, j;
    for (i = 0; i < 4; i++){
        for (j = 0; j < 4; j++){
            printf("%.5f ", L(i, j));
        }
        printf("\n");
    }
}

void tnsCopyMatrix44d(tnsMatrix44d from, tnsMatrix44d to){
    memcpy(to, from, sizeof(tnsMatrix44d));
}
void tnsMultiply44d(tnsMatrix44d result, tnsMatrix44d l, tnsMatrix44d r){
    int i;
    for (i = 0; i < 4; i++){
        real ai0 = L(i, 0), ai1 = L(i, 1), ai2 = L(i, 2), ai3 = L(i, 3);
        P(i, 0) = ai0 * R(0, 0) + ai1 * R(1, 0) + ai2 * R(2, 0) + ai3 * R(3, 0);
        P(i, 1) = ai0 * R(0, 1) + ai1 * R(1, 1) + ai2 * R(2, 1) + ai3 * R(3, 1);
        P(i, 2) = ai0 * R(0, 2) + ai1 * R(1, 2) + ai2 * R(2, 2) + ai3 * R(3, 2);
        P(i, 3) = ai0 * R(0, 3) + ai1 * R(1, 3) + ai2 * R(2, 3) + ai3 * R(3, 3);
    }
};
void tnsInverse44d(tnsMatrix44d inverse, tnsMatrix44d mat){
    int i, j, k;
    double temp;
    tnsMatrix44d tempmat;
    real max;
    int maxj;

    tnsLoadIdentity44d(inverse);

    tnsCopyMatrix44d(mat, tempmat);

    for (i = 0; i < 4; i++){
        /* Look for row with max pivot */
        max = fabsf(tempmat[i * 5]);
        maxj = i;
        for (j = i + 1; j < 4; j++){
            if (fabsf(tempmat[j * 4 + i]) > max){
                max = fabsf(tempmat[j * 4 + i]);
                maxj = j;
            }
        }

        /* Swap rows if necessary */
        if (maxj != i){
            for (k = 0; k < 4; k++){
                real t;
                t = tempmat[i * 4 + k];
                tempmat[i * 4 + k] = tempmat[maxj * 4 + k];
                tempmat[maxj * 4 + k] = t;

                t = inverse[i * 4 + k];
                inverse[i * 4 + k] = inverse[maxj * 4 + k];
                inverse[maxj * 4 + k] = t;
            }
        }

        //if (UNLIKELY(tempmat[i][i] == 0.0f)) {
        //	return false;  /* No non-zero pivot */
        //}

        temp = (double)tempmat[i * 5];
        for (k = 0; k < 4; k++){
            tempmat[i * 4 + k] = (real)((double)tempmat[i * 4 + k] / temp);
            inverse[i * 4 + k] = (real)((double)inverse[i * 4 + k] / temp);
        }
        for (j = 0; j < 4; j++){
            if (j != i){
                temp = tempmat[j * 4 + i];
                for (k = 0; k < 4; k++){
                    tempmat[j * 4 + k] -= (real)((double)tempmat[i * 4 + k] * temp);
                    inverse[j * 4 + k] -= (real)((double)inverse[i * 4 + k] * temp);
                }
            }
        }
    }
}
void tnsMakeTranslationMatrix44d(tnsMatrix44d mTrans, real x, real y, real z){
    tnsLoadIdentity44d(mTrans);
    mTrans[12] = x;
    mTrans[13] = y;
    mTrans[14] = z;
}
void tnsMakePerspectiveMatrix44d(tnsMatrix44d mProjection, real fFov_rad, real fAspect, real zMin, real zMax){
    real yMax = zMin * tanf(fFov_rad * 0.5f);
    real yMin = -yMax;
    real xMin = yMin * fAspect;
    real xMax = -xMin;

    tnsLoadIdentity44d(mProjection);

    mProjection[0] = (2.0f * zMin) / (xMax - xMin);
    mProjection[5] = (2.0f * zMin) / (yMax - yMin);
    mProjection[8] = (xMax + xMin) / (xMax - xMin);
    mProjection[9] = (yMax + yMin) / (yMax - yMin);
    mProjection[10] = -((zMax + zMin) / (zMax - zMin));
    mProjection[11] = -1.0f;
    mProjection[14] = -((2.0f * (zMax * zMin)) / (zMax - zMin));
    mProjection[15] = 0.0f;
}
void tnsMakeZTrackingMatrix44d(tnsMatrix44d mat, tnsVector3d this, tnsVector3d that, tnsVector3d up){
    tnsVector4d fwd, l, t, rt;
    fwd[3] = l[3] = t[3] = rt[3] = 1;
    t[0] = up[0];
    t[1] = up[1];
    t[2] = up[2];
    fwd[0] = that[0] - this[0];
    fwd[1] = that[1] - this[1];
    fwd[2] = that[2] - this[2];
    tnsNormalizeSelf3d(fwd);
    tnsVectorCross3d(l, fwd, t);
    tnsNormalizeSelf3d(l);
    tnsVectorCross3d(rt, l, fwd);
    tnsNormalizeSelf3d(rt);

    tnsLoadIdentity44d(mat);

    mat[0] = l[0];
    mat[1] = l[1];
    mat[2] = l[2];

    mat[4] = rt[0];
    mat[5] = rt[1];
    mat[6] = rt[2];

    mat[8] = -fwd[0];
    mat[9] = -fwd[1];
    mat[10] = -fwd[2];
}
void tnsMakeZTrackingMatrixDelta44d(tnsMatrix44d mat, tnsVector3d delta, tnsVector3d up){
    tnsVector4d fwd, l, t, rt;
    fwd[3] = l[3] = t[3] = rt[3] = 1;
    t[0] = up[0];
    t[1] = up[1];
    t[2] = up[2];
    fwd[0] = delta[0];
    fwd[1] = delta[1];
    fwd[2] = delta[2];

    tnsLoadIdentity44d(mat);

    tnsVectorCross3d(l, t, fwd);
    tnsVectorCross3d(rt, fwd, l);

    tnsNormalizeSelf3d(l);
    tnsNormalizeSelf3d(rt);
    tnsNormalizeSelf3d(fwd);

    mat[0] = l[0];
    mat[1] = l[1];
    mat[2] = l[2];

    mat[4] = rt[0];
    mat[5] = rt[1];
    mat[6] = rt[2];

    mat[8] = fwd[0];
    mat[9] = fwd[1];
    mat[10] = fwd[2];
}
void tnsMakeOrthoMatrix44d(tnsMatrix44d mProjection, real xMin, real xMax, real yMin, real yMax, real zMin, real zMax){
    tnsLoadIdentity44d(mProjection);

    mProjection[0] = 2.0f / (xMax - xMin);
    mProjection[5] = 2.0f / (yMax - yMin);
    mProjection[10] = -2.0f / (zMax - zMin);
    mProjection[12] = -((xMax + xMin) / (xMax - xMin));
    mProjection[13] = -((yMax + yMin) / (yMax - yMin));
    mProjection[14] = -((zMax + zMin) / (zMax - zMin));
    mProjection[15] = 1.0f;
}
void tnsMakeRotationMatrix44d(tnsMatrix44d m, real angle_rad, real x_, real y_, real z_){
    float x = x_;
    float y = y_;
    float z = z_;

    float c = cos(angle_rad);
    float s = sin(angle_rad);
    tnsMatrix44d d={
       x*x*(1.0f-c)+c,   x*y*(1.0f-c)-z*s, x*z*(1.0f-c)+y*s, 0.0f,
       y*x*(1.0f-c)+z*s, y*y*(1.0f-c)+c,   y*z*(1.0f-c)-x*s, 0.0f,
       z*x*(1.0f-c)-y*s, z*y*(1.0f-c)+x*s, z*z*(1.0f-c)+c,   0.0f,
       0.0f,             0.0f,             0.0f,             1.0f };
    memcpy(m,d,sizeof(tnsMatrix44d));
}
void tnsMakeRotationXMatrix44d(tnsMatrix44d m, real angle_rad){
    tnsLoadIdentity44d(m);
    m[5] = cos(angle_rad);
    m[6] = sin(angle_rad);
    m[9] = -sin(angle_rad);
    m[10] = cos(angle_rad);
}
void tnsMakeRotationYMatrix44d(tnsMatrix44d m, real angle_rad){
    tnsLoadIdentity44d(m);
    m[0] = cos(angle_rad);
    m[2] = -sin(angle_rad);
    m[8] = sin(angle_rad);
    m[10] = cos(angle_rad);
}
void tnsMakeRotationZMatrix44d(tnsMatrix44d m, real angle_rad){
    tnsLoadIdentity44d(m);
    m[0] = cos(angle_rad);
    m[1] = sin(angle_rad);
    m[4] = -sin(angle_rad);
    m[5] = cos(angle_rad);
}
void tnsMakeScaleMatrix44d(tnsMatrix44d m, real x, real y, real z){
    tnsLoadIdentity44d(m);
    m[0] = x;
    m[5] = y;
    m[10] = z;
}
void tnsMakeViewportMatrix44d(tnsMatrix44d m, real w, real h, real Far, real Near){
    tnsLoadIdentity44d(m);
    m[0] = w / 2;
    m[5] = h / 2;
    m[10] = (Far - Near) / 2;
    m[12] = w / 2;
    m[13] = h / 2;
    m[14] = (Far + Near) / 2;
    m[15] = 1;
    //m[0] = 2/w;
    //m[5] = 2/h;
    //m[10] = 1;
    //m[12] = 2/w;
    //m[13] = 2/h;
    //m[14] = 1;
    //m[15] = 1;
}

void tnsInitFirstLevel(tnsMatrixStack *tms){
    tnsLoadIdentity44d(tms->level[0].model);
    tnsLoadIdentity44d(tms->level[0].projection);
    tnsLoadIdentity44d(tms->level[0].view);
}

//-------------------Export

void tnsOrtho(real xMin, real xMax, real yMin, real yMax, real zMin, real zMax){
    tnsShader *current_shader = 0;
    real *mat = tnsGetProjectionMatrix();

    tnsMakeOrthoMatrix44d(mat, xMin, xMax, yMin, yMax, zMin, zMax);

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyProjection(current_shader, mat);
    }
}
void tnsPerspective(real fFov_rad, real fAspect, real zMin, real zMax){
    tnsShader *current_shader = 0;
    real *mat = tnsGetProjectionMatrix();

    tnsMakePerspectiveMatrix44d(mat, fFov_rad, fAspect, zMin, zMax);

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyProjection(current_shader, mat);
    }
}

void tnsTranslate3d(real x, real y, real z){
    tnsShader *current_shader = 0;
    real *mat = tnsGetModelMatrix();
    tnsMatrix44d transmat, result;

    tnsMakeTranslationMatrix44d(transmat, x, y, z);
    tnsMultiply44d(result, mat, transmat);
    memcpy(mat, result, sizeof(tnsMatrix44d));

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyModel(current_shader, result);
    }
}
void tnsPreTranslate3d(real x, real y, real z){
    tnsShader *current_shader = 0;
    real *mat = tnsGetModelMatrix();
    tnsMatrix44d transmat, result;

    //glTranslatef(x, y, z);

    tnsMakeTranslationMatrix44d(transmat, x, y, z);
    tnsMultiply44d(result, mat, transmat);
    memcpy(mat, result, sizeof(tnsMatrix44d));
}
void tnsRotate4d(real degrees, real x, real y, real z){
    tnsShader *current_shader = 0;
    real *mat = tnsGetModelMatrix();
    tnsMatrix44d rotmat, result;

    tnsMakeRotationMatrix44d(rotmat, rad(degrees), x, y, z);
    tnsMultiply44d(result, mat, rotmat);
    memcpy(mat, result, sizeof(tnsMatrix44d));

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyModel(current_shader, result);
    }
}
void tnsPreRotate4d(real degrees, real x, real y, real z){
    tnsShader *current_shader = 0;
    real *mat = tnsGetModelMatrix();
    tnsMatrix44d rotmat, result;

    tnsMakeRotationMatrix44d(rotmat, rad(degrees), x, y, z);
    tnsMultiply44d(result, mat, rotmat);
    memcpy(mat, result, sizeof(tnsMatrix44d));
}
void tnsScale3d(real x, real y, real z){
    tnsShader *current_shader = 0;
    real *mat = tnsGetModelMatrix();
    tnsMatrix44d scalemat, normal_scaler, result;

    glScalef(x, y, z);

    tnsMakeScaleMatrix44d(scalemat, x, y, z);

    tnsMultiply44d(result, mat, scalemat);
    memcpy(mat, result, sizeof(tnsMatrix44d));

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyModel(current_shader, result);
    }
}
void tnsPreScale3d(real x, real y, real z){
    tnsShader *current_shader = 0;
    real *mat = tnsGetModelMatrix();
    tnsMatrix44d scalemat, normal_scaler, result;

    glScalef(x, y, z);

    tnsMakeScaleMatrix44d(scalemat, x, y, z);

    tnsMultiply44d(result, mat, scalemat);
    memcpy(mat, result, sizeof(tnsMatrix44d));
}

void tKnlPopMatrix();
void tKnlPushMatrix();
void tnsPushMatrix(){
    tKnlPushMatrix();
};
void tnsPopMatrix(){
    tKnlPopMatrix();
}

//========================================[KNL]

void tnsInitRenderKernel(int matrixStackLevel){
    tnsCommand *c;
    T = memAcquireHyper(sizeof(tnsMain));
    GLuint m_nQuadVAO;

    T->GLVersionStr = glGetString(GL_VERSION);
    T->GLVendorStr = glGetString(GL_VENDOR);
    T->GLRendererStr = glGetString(GL_RENDERER);
    T->GLSLVersionStr = glGetString(GL_SHADING_LANGUAGE_VERSION);

    T->stack.max_level = matrixStackLevel;
    T->stack.level = CreateNewBuffer(tnsMatrixStackItem, matrixStackLevel);
    T->NextShaderIndex = 1;
    T->BindedShader = -1;
    tnsInitFirstLevel(&T->stack);

    /* Needed for gl3.3+ */
    glGenVertexArrays(1,&T->GlobalVAO);
    glBindVertexArray(T->GlobalVAO);

    arrEnsureLength(&T->Vert, T->NextVert, &T->MaxVert, sizeof(GLfloat));
    glGenBuffers(1, &T->VertBufObject);
    glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
    glBufferData(GL_ARRAY_BUFFER, T->MaxVert * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);

    arrEnsureLength(&T->Color, T->NextColor, &T->MaxColor, sizeof(GLfloat));
    glGenBuffers(1, &T->ColorBufObject);
    glBindBuffer(GL_ARRAY_BUFFER, T->ColorBufObject);
    glBufferData(GL_ARRAY_BUFFER, T->MaxColor * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);

    arrEnsureLength(&T->Normal, T->NextNormal, &T->MaxNormal, sizeof(GLfloat));
    glGenBuffers(1, &T->NormalBufObject);
    glBindBuffer(GL_ARRAY_BUFFER, T->NormalBufObject);
    glBufferData(GL_ARRAY_BUFFER, T->MaxNormal * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);

    arrEnsureLength(&T->TexCoord, T->NextTexCoord, &T->MaxTexCoord, sizeof(GLfloat));
    glGenBuffers(1, &T->TexCoordBufObject);
    glBindBuffer(GL_ARRAY_BUFFER, T->TexCoordBufObject);
    glBufferData(GL_ARRAY_BUFFER, T->MaxTexCoord * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);

    arrEnsureLength(&T->Index, T->NextIndex, &T->MaxIndex, sizeof(GLuint));
    glGenBuffers(1, &T->IndexBufObject);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, T->IndexBufObject);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, T->MaxIndex * sizeof(GLuint), 0, GL_DYNAMIC_DRAW);

    arrEnsureLength(&T->DrawingCommand, T->NextCommand, &T->MaxCommand, sizeof(tnsCommand));
}
void tnsInitBuiltinShaders(){
    T->immShader = tnsNewShaderProgram(
        tnsNewVertexShader(LA_IMM_VERTEX_SHADER),tnsNewFragmentShader(LA_IMM_FRAGMENT_SHADER),-1);

    T->TEST_MatcapShader = tnsNewShaderProgram(
        tnsNewVertexShader(TNS_VERTEX_SIMPLE_MATCAP),
        tnsNewFragmentShader(TNS_FRAGMENT_SIMPLE_MATCAP), -1);

    T->TransparentGridShader = tnsNewShaderProgram(
        tnsNewVertexShader(TNS_VERTEX_GRID),
        tnsNewFragmentShader(TNS_FRAGMENT_TRANSPARNT_GRID), -1);

    T->RayShader = tnsNewShaderProgram(
        tnsNewVertexShader(LA_RAY_VERTEX_SHADER),
        tnsNewFragmentShader(LA_RAY_FRAGMENT_SHADER), -1);

    T->ShadowShader = tnsNewShaderProgram(
        tnsNewVertexShader(LA_CASCADE_SHADOW_VERTEX_SHADER),
        tnsNewFragmentShader(LA_CASCADE_SHADOW_FRAGMENT_SHADER), -1);

    T->SceneShader = tnsNewShaderProgram(
        tnsNewVertexShader(LA_SCENE_VERTEX_SHADER),
        tnsNewFragmentShader(LA_SCENE_FRAGMENT_SHADER), -1);

    T->SelectionShader = tnsNewShaderProgram(
        tnsNewVertexShader(LA_SELECTION_VERTEX_SHADER),
        tnsNewFragmentShader(LA_SELECTION_FRAGMENT_SHADER), -1);

    tnsUseShader(T->immShader);
    tnsEnableShaderv(T->immShader);
}
void tnsInitWindowDefaultRenderConfig(){
    tnsInitBuiltinShaders();
};
void tnsDeleteBuiltinShaders(){
    tnsDeleteShaderProgram(T->immShader); T->immShader=0;
    tnsDeleteShaderProgram(T->TEST_MatcapShader); T->TEST_MatcapShader=0;
    tnsDeleteShaderProgram(T->TransparentGridShader); T->TransparentGridShader=0;
    tnsDeleteShaderProgram(T->SelectionShader); T->SelectionShader=0;
    tnsDeleteShaderProgram(T->SceneShader); T->SceneShader=0;
    tnsDeleteShaderProgram(T->ShadowShader); T->ShadowShader=0;
    tnsDeleteShaderProgram(T->RayShader); T->RayShader=0;
}
void tnsQuit(){
    tnsQuitFontManager();

    tnsDeleteBuiltinShaders();

    tnsOffscreen* off; while(off=T->Offscreens.pFirst){ tnsDelete2DOffscreen(off); }
    tnsTexture* t; while(t=T->Textures.pFirst){ tnsDeleteTexture(t); }
    
    arrFree(&T->Vert, &T->MaxVert);
    arrFree(&T->Color, &T->MaxColor);
    arrFree(&T->Normal, &T->MaxNormal);
    arrFree(&T->TexCoord, &T->MaxTexCoord);
    arrFree(&T->Index, &T->MaxIndex);
    arrFree(&T->DrawingCommand, &T->DrawingCommand);

    glDeleteBuffers(1, &T->VertBufObject);
    glDeleteBuffers(1, &T->ColorBufObject);
    glDeleteBuffers(1, &T->NormalBufObject);
    glDeleteBuffers(1, &T->TexCoordBufObject);
    glDeleteBuffers(1, &T->IndexBufObject);

    glBindVertexArray(0);
    glDeleteVertexArrays(1,&T->GlobalVAO);

    FreeMem(T->stack.level);
    memFree(T);
}

tnsMatrixStackItem *tKnlGetCurrentMatStackItem(){
    return &T->stack.level[T->stack.current_level];
}
tnsShader *tKnlGetActiveShader(){
    return T->CurrentShader;
}
real *tnsGetModelMatrix(){
    return tKnlGetCurrentMatStackItem()->model;
}
real *tnsGetViewMatrix(){
    return tKnlGetCurrentMatStackItem()->view;
}
real *tnsGetProjectionMatrix(){
    return tKnlGetCurrentMatStackItem()->projection;
}
void tnsGetMVMatrix(tnsMatrix44d r){
    tnsMatrixStackItem *tmsi=tKnlGetCurrentMatStackItem();
    tnsMatrix44d tmp;
    tnsMultiply44d(r, tmsi->model, tmsi->view); 
}
void tnsGetMVPMatrix(tnsMatrix44d r){
    tnsMatrixStackItem *tmsi=tKnlGetCurrentMatStackItem();
    tnsMatrix44d tmp;
    tnsMultiply44d(tmp, tmsi->model, tmsi->view); 
    tnsMultiply44d(r, tmp, tmsi->projection); 
}
void tnsResetModelMatrix(){
    tnsLoadIdentity44d(tnsGetModelMatrix());
}
void tnsResetViewMatrix(){
    tnsLoadIdentity44d(tnsGetViewMatrix());
}
void tnsResetProjectionMatrix(){
    tnsLoadIdentity44d(tnsGetProjectionMatrix());
}
int tKnlAttatchShader(tnsShader *tns){
    lstAppendItem(&T->Shaders, tns);
    tns->CustomID = T->NextShaderIndex;
    T->NextShaderIndex++;
    return tns->CustomID;
};
int CMP_NUM_IsThisShader(tnsShader *tns, int *index){
    return (tns->CustomID == *index);
}
tnsShader *tKnlFindShader1i(int CustomIndex){
    return lstFindItem(&CustomIndex, CMP_NUM_IsThisShader, &T->Shaders);
}
void tKnlPushMatrix(){
    tnsMatrixStack *tms = &T->stack;
    tnsShader *current_shader = tKnlGetActiveShader();

    if (tms->current_level == tms->max_level || !current_shader) return;

    memcpy(&tms->level[tms->current_level + 1], &tms->level[tms->current_level], sizeof(tnsMatrixStackItem));

    tms->current_level++;

    //tnsShaderApplyModel(current_shader,tms->level[tms->current_level].model);
    //tnsShaderApplyView(current_shader,tms->level[tms->current_level].view);
    //tnsShaderApplyProjection(current_shader,tms->level[tms->current_level].projection);
};
void tKnlPopMatrix(){
    tnsMatrixStack *tms = &T->stack;
    tnsShader *current_shader = tKnlGetActiveShader();

    if (tms->current_level == 0 || !current_shader) return;

    tms->current_level--;

    tnsShaderApplyModel(current_shader, tms->level[tms->current_level].model);
    tnsShaderApplyView(current_shader, tms->level[tms->current_level].view);
    tnsShaderApplyProjection(current_shader, tms->level[tms->current_level].projection);
}
laListHandle *tKnlGetTextureList(){
    return &T->Textures;
}

tnsBatch *tnsCreateBatch(u32bit NumVert, int Dimension, float *Data, int NormalDimension, float *Normal, int ColorDimension, float *Colors){
    tnsBatch *b = CreateNew(tnsBatch);
    b->Dimension = Dimension;
    b->NormalDimension=NormalDimension;
    b->ColorDimension=ColorDimension;
    b->NumVert = NumVert;

    glGenBuffers(1, &b->VBO);
    glBindBuffer(GL_ARRAY_BUFFER, b->VBO);
    glBufferData(GL_ARRAY_BUFFER, NumVert * Dimension * sizeof(GLfloat), Data, GL_DYNAMIC_DRAW);

    if (Normal){
        glGenBuffers(1, &b->NBO); glBindBuffer(GL_ARRAY_BUFFER, b->NBO);
        glBufferData(GL_ARRAY_BUFFER, NumVert * NormalDimension * sizeof(GLfloat), Normal, GL_DYNAMIC_DRAW);
        b->HasNormal=1;
    }
    if (Colors){
        glGenBuffers(1, &b->CBO); glBindBuffer(GL_ARRAY_BUFFER, b->CBO);
        glBufferData(GL_ARRAY_BUFFER, NumVert * ColorDimension * sizeof(GLfloat), Colors, GL_DYNAMIC_DRAW);
        b->HasColor=1;
    }

    return b;
}
tnsBatch *tnsCreateBatchi(u32bit NumVert, int Dimension, int *Data){
    tnsBatch *b = CreateNew(tnsBatch);
    b->Dimension = Dimension;
    b->NumVert = NumVert;

    glGenBuffers(1, &b->VBO);
    glBindBuffer(GL_ARRAY_BUFFER, b->VBO);
    glBufferData(GL_ARRAY_BUFFER, NumVert * Dimension * sizeof(GLuint), Data, GL_DYNAMIC_DRAW);

    return b;
}
void tnsCommandUseUniformColor(tnsBatchCommand*c, real* color){
    tnsVectorCopy4d(color, c->UniformColor);
    c->UseUniformColor=1;
}
void tnsCommandOverrideColorArray(tnsBatchCommand*c, int VertCount, int ColorDimension, float* colors){
    if(!colors) return;
    c->OverrideColorArray=1; 
    glGenBuffers(1, &c->CBO); glBindBuffer(GL_ARRAY_BUFFER, c->CBO);
    glBufferData(GL_ARRAY_BUFFER, VertCount* ColorDimension * sizeof(GLfloat), colors, GL_DYNAMIC_DRAW);
    c->ColorDimension = ColorDimension;
}
tnsBatchCommand *tnsCreateCommand(tnsBatch *b, const char* name, u32bit ElementCount, int Dimension, GLenum DrawAs, u32bit *Elements, int HiddenByDefault){
    tnsBatchCommand *bc = CreateNew(tnsBatchCommand);
    bc->ElementCount = ElementCount;
    bc->DrawAs = DrawAs;
    bc->Dimension = Dimension;
    bc->name=name;
    bc->HiddenByDefault = HiddenByDefault;

    if(Elements){
        glGenBuffers(1, &bc->EBO); glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bc->EBO);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, ElementCount * Dimension * sizeof(GLuint), Elements, GL_DYNAMIC_DRAW);
        bc->DrawElements=1;
    }else{ bc->EBO=-1; }
    //int a = GetLastError();
    lstAppendItem(&b->Branches, bc);

    return bc;
}
void tnsDeleteBatch(tnsBatch *b){
    tnsBatchCommand *bc, *NextBc;

    glDeleteBuffers(1, &b->VBO);
    if (b->NBO > -1) glDeleteBuffers(1, &b->NBO);
    if (b->CBO > -1) glDeleteBuffers(1, &b->CBO);

    for (bc = b->Branches.pFirst; bc; bc = NextBc){
        NextBc = bc->Item.pNext;
        lstRemoveItem(&b->Branches, bc);
        if (bc->EBO > -1) glDeleteBuffers(1, &bc->EBO);
        if (bc->CBO > -1) glDeleteBuffers(1, &bc->CBO);
        FreeMem(bc);
    }

    FreeMem(b);
}
void tnsDrawBatch(tnsBatch* batch, const char* OverrideCommand, real* OverrideUniformColor, int OverrideAsArray) {
	//int Mode = batch->DrawMode;
	if (!batch) return;
    tnsShader* cs=T->BindedShader;

	//glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

	glBindBuffer(GL_ARRAY_BUFFER, batch->VBO);
	glEnableVertexAttribArray(cs->iVertex);
	glVertexAttribPointer(cs->iVertex, batch->Dimension, GL_FLOAT, 0, 0, 0);

    if(cs->iNormal>=0){
        if(batch->HasNormal){
            glBindBuffer(GL_ARRAY_BUFFER, batch->NBO); glEnableVertexAttribArray(cs->iNormal);
            glVertexAttribPointer(cs->iNormal, batch->NormalDimension, GL_FLOAT, 0, 0, 0);
        }else{ glDisableVertexAttribArray(cs->iNormal); glVertexAttrib3f(cs->iNormal,0,0,1); }
    }
    if(cs->iColor>=0){
        if(batch->HasColor){
            glBindBuffer(GL_ARRAY_BUFFER, batch->CBO); glEnableVertexAttribArray(cs->iColor);
            glVertexAttribPointer(cs->iColor, batch->ColorDimension, GL_FLOAT, 0, 0, 0);
        }
    }

    int IsOverrideColor=0;
	for (tnsBatchCommand* bc = batch->Branches.pFirst; bc; bc = bc->Item.pNext) {
        if(OverrideCommand && !strSame(OverrideCommand, bc->name)){ continue; }
        if(!OverrideCommand && bc->HiddenByDefault){ continue; }
        if(cs->iColor){
            if(bc->OverrideColorArray){
                glBindBuffer(GL_ARRAY_BUFFER, bc->CBO); glEnableVertexAttribArray(cs->iColor);
                glVertexAttribPointer(cs->iColor, bc->ColorDimension, GL_FLOAT, 0, 0, 0); IsOverrideColor=1;
            }else{
                if(batch->HasColor){ if(IsOverrideColor){ glBindBuffer(GL_ARRAY_BUFFER, batch->CBO); glEnableVertexAttribArray(cs->iColor);
                    glVertexAttribPointer(cs->iColor, batch->ColorDimension, GL_FLOAT, 0, 0, 0); }
                }else{ glDisableVertexAttribArray(cs->iColor); glVertexAttrib4fv(cs->iColor,T->StateColor); }
            }
            if(bc->UseUniformColor || OverrideUniformColor){ glDisableVertexAttribArray(cs->iColor); IsOverrideColor=1;
                if(OverrideUniformColor){ glVertexAttrib4f(cs->iColor,LA_COLOR4(OverrideUniformColor)); }
                else glVertexAttrib4f(cs->iColor,LA_COLOR4(bc->UniformColor)); 
            }
        }
        int DrawElements=OverrideAsArray?0:bc->DrawElements;
        if(DrawElements){
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bc->EBO);
            glDrawElements(bc->DrawAs, bc->ElementCount*bc->Dimension, GL_UNSIGNED_INT, 0);
        }else{ glDrawArrays(bc->DrawAs,0,bc->ElementCount); }
	}

	//glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

//==========================*=======================[Texture]

int tnsGetTextureMemoryComponetCount(tnsTexture *t){
    int Comp = 0;
    int CompSize;
    switch (t->GLTexBitsType){
    case GL_RGB:
        Comp = 3;
        break;
    case GL_RGBA:
        Comp = 4;
        break;
    }
    return t->Width * t->Height * Comp;
}

int tnsInit2DTexture(tnsTexture *t, GLint glInternalFormat, int w, int h, int Multisample){
    if (!t) return 0;

    if(0/*glInternalFormat==GL_DEPTH_COMPONENT32F*/){
        t->Width = w;
        t->Height = h;
        t->GLTexBitsType = GL_DEPTH_COMPONENT32F;
        t->GLTexType = GL_RENDERBUFFER;
        t->Multisample = Multisample;
        glGenRenderbuffers(1, &t->GLTexHandle);
        
        tnsConfigure2DTexture(t);
    }else{
        t->Width = w;
        t->Height = h;
        t->GLTexBitsType = glInternalFormat;
        t->GLTexType = Multisample?GL_TEXTURE_2D_MULTISAMPLE:GL_TEXTURE_2D;
        t->Multisample = Multisample;
        glGenTextures(1, &t->GLTexHandle);

        tnsConfigure2DTexture(t);
    }

    return 1;
}
int tnsInit3DTexture(tnsTexture *t, GLint glInternalFormat, int w, int h, int slices){
    if (!t) return 0;
    t->Width = w;
    t->Height = h;
    t->GLTexBitsType = glInternalFormat;
    t->GLTexType = GL_TEXTURE_3D;
    glGenTextures(1, &t->GLTexHandle);
    tnsConfigure3DTexture(t);
    return 1;
}

void tnsConfigure2DTexture(tnsTexture *t){
    tnsBindTexture(t);
    if(t->GLTexType==GL_RENDERBUFFER){
        if(t->Multisample){ glRenderbufferStorageMultisample(GL_RENDERBUFFER, t->Multisample, t->GLTexBitsType, t->Width, t->Height); }
        else { glRenderbufferStorage(GL_RENDERBUFFER, t->GLTexBitsType, t->Width, t->Height); }
    }else{
        int isDepth=t->GLTexBitsType==GL_DEPTH_COMPONENT||t->GLTexBitsType==GL_DEPTH_COMPONENT16||
            t->GLTexBitsType==GL_DEPTH_COMPONENT24||t->GLTexBitsType==GL_DEPTH_COMPONENT32F;
        int format=isDepth?GL_DEPTH_COMPONENT:GL_RGBA;
        if(t->Multisample) glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, t->Multisample, t->GLTexBitsType, t->Width, t->Height, GL_TRUE);
        else{ glTexImage2D(GL_TEXTURE_2D, 0, t->GLTexBitsType, t->Width, t->Height, 0, format, GL_UNSIGNED_BYTE, 0);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                //glTexEnvi(GL_TEXTURE_2D, GL_TEXTURE_ENV_MODE, GL_REPLACE);
        }
    }
    tnsUnbindTexture(t);
}
void tnsConfigure3DTexture(tnsTexture *t){
    tnsBindTexture(t);
    glTexImage3D(GL_TEXTURE_3D, 0, t->GLTexBitsType, t->Width, t->Height, t->Slices, 0, GL_RGBA, GL_FLOAT, 0);
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        //glTexEnvi(GL_TEXTURE_2D, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    tnsUnbindTexture(t);
}

void tnsReconfigureTextureParameters(int Multisample){
    laListHandle* l = tKnlGetTextureList();
    for(tnsOffscreen* o=T->Offscreens.pFirst;o;o=o->Item.pNext){
        tnsTexture* t=o->pColor[0];
        if(t){
            t->Multisample = Multisample;
            int recreate=0;
            if(t->Multisample){ if(t->GLTexType==GL_TEXTURE_2D){t->GLTexType=GL_TEXTURE_2D_MULTISAMPLE; recreate=1;}}
            else { if(t->GLTexType==GL_TEXTURE_2D_MULTISAMPLE){t->GLTexType=GL_TEXTURE_2D; recreate=1;}}
            if(recreate){
                glDeleteTextures(1, &t->GLTexHandle);
                glGenTextures(1, &t->GLTexHandle);
                T->TexColor=0;
            }
            tnsConfigure2DTexture(t);
            tnsAttach2DOffscreenBuffer(o, GL_COLOR_ATTACHMENT0, t);
        }
        t=o->pDepth;
        if(t){
            t->Multisample = Multisample;
            tnsConfigure2DTexture(t);
            tnsAttach2DOffscreenBuffer(o, GL_DEPTH_ATTACHMENT, t);
        }
    }
}

tnsTexture *tnsCreate2DTexture(GLint glInternalFormat, int w, int h, int Multisample){
    tnsTexture *tex = CreateNew(tnsTexture);
    tnsInit2DTexture(tex, glInternalFormat, w, h, Multisample);
    laNotifyUsers("tns.texture_list");
    lstAppendItem(tKnlGetTextureList(), tex);
    return tex;
};
tnsTexture *tnsCreate3DTexture(GLint glInternalFormat, int w, int h, int slices){
    tnsTexture *tex = CreateNew(tnsTexture);
    tnsInit3DTexture(tex, glInternalFormat, w, h, slices);
    laNotifyUsers("tns.texture_list");
    lstAppendItem(tKnlGetTextureList(), tex);
    return tex;
};
void tnsCopyScreenTo2DTexture(tnsTexture *target, int x_lower_left, int y_lower_left, int w, int h){
    if(target->GLTexType!=GL_TEXTURE_2D) return;
    tnsBindTexture(target);
    glReadBuffer(GL_BACK);
    glCopyTexSubImage2D(target->GLTexType, 0, 0, 0, x_lower_left, y_lower_left, w, h);
    tnsUnbindTexture();
}

void tnsUseMaskTexture(tnsTexture *t){
    if(!t){T->StateTextureMode=0; return;}
    T->StateTexColor = t; T->StateTextureMode=1;
}
void tnsUseTexture(tnsTexture *t){
    if(!t){T->StateTextureMode=0; return;}
    if(t->GLTexType == GL_TEXTURE_2D){ T->StateTexColor = t; T->StateTextureMode=2; }
    else if(t->GLTexType == GL_TEXTURE_2D_MULTISAMPLE){ T->StateTexColor = t; T->StateTextureMode=3; }
}
void tnsUseNoTexture(){
    tnsUseTexture(0);
}
void tnsUseMultiplyColor(int enable){
    T->StateMultiplyColor=enable?1:0;
}

void tnsActiveTexture(GLenum tex){
    if (T->GlTextureSets != tex) glActiveTexture(tex);
    T->GlTextureSets = tex;
}
void tnsBindTexture(tnsTexture *t){
    if (!t || T->TexColor==t) return;
    if(t->GLTexType == GL_TEXTURE_2D){ tnsActiveTexture(GL_TEXTURE0); glBindTexture(t->GLTexType, t->GLTexHandle); T->TexColor=t;}
    elif(t->GLTexType == GL_TEXTURE_2D_MULTISAMPLE){ tnsActiveTexture(GL_TEXTURE1); glBindTexture(t->GLTexType, t->GLTexHandle); T->TexColor=t;}
    elif(t->GLTexType == GL_RENDERBUFFER){ glBindRenderbufferEXT(GL_RENDERBUFFER, t->GLTexHandle); T->TexRenderbuffer = t;}
    elif(t->GLTexType == GL_TEXTURE_3D){ tnsActiveTexture(GL_TEXTURE0); glBindTexture(t->GLTexType, t->GLTexHandle); T->TexColor=t;}
}
void tnsUnbindTexture(){
    if(T->TexRenderbuffer){ if(T->TexRenderbuffer->GLTexType == GL_RENDERBUFFER){ glBindRenderbufferEXT(GL_RENDERBUFFER, 0); T->TexRenderbuffer=0;}}
    if(T->TexColor){
        if(T->TexColor->GLTexType == GL_TEXTURE_2D){tnsActiveTexture(GL_TEXTURE0);}
        else if(T->TexColor->GLTexType == GL_TEXTURE_2D_MULTISAMPLE){tnsActiveTexture(GL_TEXTURE1);}
        else if(T->TexColor->GLTexType == GL_TEXTURE_3D){tnsActiveTexture(GL_TEXTURE0);}
        glBindTexture(T->TexColor->GLTexType, 0); T->TexColor=0;
    }
}
void tnsUniformUseTexture(tnsShader* s, int mode, int sample){
    if(s->StateTextureMode != mode){ s->StateTextureMode=mode; glUniform1i(s->iTextureMode,mode); }
    if(mode==3 && s->StateSampleAmount != sample){ s->StateSampleAmount=sample, glUniform1i(s->iSampleAmount,sample); }
}
void tnsUniformUseMultiplyColor(tnsShader* s, int enable){
    int mode=enable?1:0;
    if(s->StateMultiplyColor != mode){ s->StateMultiplyColor=mode; glUniform1i(s->iMultiplyColor,mode); }
}

void tnsDraw2DTextureDirectly(tnsTexture *t, real x, real y, real w, real h){
    real Verts[8];
    real UV[8] = {
        0.0f, 1.0f,
        1.0f, 1.0f,
        1.0f, 0.0f,
        0.0f, 0.0f};
    tnsMakeQuad2d(Verts, x, y, x + w, y, x + w, y + h, x, y + h);
    tnsUseTexture(t);
    tnsVertexArray2d(Verts, 4);
    tnsTexCoordArray2d(UV, 4);
    tnsPackAs(GL_TRIANGLE_FAN);
}
void tnsDraw2DTextureArg(tnsTexture *t,
                         real x_upper_right, real y_upper_right, int w, int h,
                         real *MultiplyColor,
                         real LPadding, real RPadding, real TPadding, real BPadding){
    real Verts[8];
    real UV[8] = {
        0.0f + LPadding, 1.0f - TPadding,
        1.0f - RPadding, 1.0f - TPadding,
        1.0f - RPadding, 0.0f + BPadding,
        0.0f + LPadding, 0.0f + BPadding};
    tnsMakeQuad2d(Verts,
                  x_upper_right, y_upper_right,
                  x_upper_right + w, y_upper_right,
                  x_upper_right + w, y_upper_right + h,
                  x_upper_right, y_upper_right + h);

    if (MultiplyColor){ tnsColor4dv(MultiplyColor); tnsUseMultiplyColor(1); }

    tnsUseTexture(t);
    tnsVertexArray2d(Verts, 4);
    tnsTexCoordArray2d(UV, 4);
    tnsPackAs(GL_TRIANGLE_FAN);
    tnsUseMultiplyColor(0);
}
void tnsDeleteTexture(tnsTexture *t){
    laListHandle *lst = tKnlGetTextureList();

    tnsUnbindTexture();

    if (!t) return;

    if (t->GLTexType == GL_RENDERBUFFER){
        glBindRenderbufferEXT(GL_RENDERBUFFER, t->GLTexHandle);
        glDeleteRenderbuffersEXT(1, &t->GLTexHandle);
    }else{
        //glBindTexture(t->GLTexType, 0);
        glDeleteTextures(1, &t->GLTexHandle);
    }
    laNotifyUsers("tns.texture_list");
    lstRemoveItem(lst, t);
    if (t->DrawData) FreeMem(t->DrawData);
    FreeMem(t);
}

int tnsTextureMemorySize(tnsTexture *t, int Mem){
    int ElemSize;

    if (Mem) return t->Width * t->Height * sizeof(void *);

    switch (t->GLTexBitsType){
    default:
    case GL_RED:
        ElemSize = sizeof(char) * 1;
        break;
    case GL_RG:
        ElemSize = sizeof(char) * 2;
        break;
    case GL_RGB:
        ElemSize = sizeof(char) * 3;
        break;
    case GL_RGBA:
        ElemSize = sizeof(char) * 4;
        break;
    case GL_RGBA32F:
        ElemSize = sizeof(float) * 4;
        break;
    case GL_DEPTH_COMPONENT32F:
        ElemSize = sizeof(float);
        break;
    }

    t->ElemSize = ElemSize;

    return t->Width * t->Height * ElemSize;
}

void tnsCreateTextureReadbackBuffer(tnsTexture *t){
    if (t->SamplePtr){
        tnsTextureSample *tns;
        memset(t->SamplePtr, 0, tnsTextureMemorySize(t, 1));
        while (tns = lstPopItem(&t->ErasedSamples))
            FreeMem(tns);
        while (tns = lstPopItem(&t->PendingSamples))
            FreeMem(tns);
    }else{
        t->TextureReadBack = calloc(1, tnsTextureMemorySize(t, 0));
        t->SamplePtr = calloc(1, tnsTextureMemorySize(t, 1));
    }
    t->RBWidth = t->Width;
    t->RBHeight = t->Height;
}
void tnsDeleteTextureReadbackBuffer(tnsTexture *t){
    FreeMem(t->TextureReadBack);
    FreeMem(t->SamplePtr);
    tnsTextureSample *tns;
    while (tns = lstPopItem(&t->ErasedSamples))
        FreeMem(tns);
    while (tns = lstPopItem(&t->PendingSamples))
        FreeMem(tns);
    t->RBWidth = t->RBHeight = 0;
}
void tnsReadbackTexture(tnsTexture *t){
    //if (t->RBWidth != t->Width || t->RBHeight != t->Height) tnsDeleteTextureReadbackBuffer(t);
//
    //tnsCreateTextureReadbackBuffer(t);
//
    //tnsBindTexture(t);
    //glGetTexImage(t->GLTexType, 0, t->DataFormat, t->DataType, t->TextureReadBack);
    //tnsUnbindTexture();
//
    //int w, h;
    //tnsTextureSample *tns;
    //u8bit sample;
    //for (h = 0; h < t->Height; h++){
    //    for (w = 0; w < t->Width; w++){
    //        int index = h * t->Width + w;
    //        if ((sample = *tnsTextureSampleU8(t, w, h)) > TNS_SNAKE_STRENGTH_LIMIT){
    //            tns = CreateNew(tnsTextureSample);
    //            lstAppendItem(&t->PendingSamples, tns);
    //            t->SamplePtr[h * t->Width + w] = tns;
    //            tns->X = w;
    //            tns->Y = h;
    //            tns->Sample = sample;
    //        }
    //    }
    //}
}

//====================================================[NEW RENDER KERNEL]
//=================[Immediate-style api]

void tnsColor4d(real r, real g, real b, real a){
    T->StateColor[0] = r;
    T->StateColor[1] = g;
    T->StateColor[2] = b;
    T->StateColor[3] = a;
}
void tnsColor4dv(real *rgba){
    T->StateColor[0] = rgba[0];
    T->StateColor[1] = rgba[1];
    T->StateColor[2] = rgba[2];
    T->StateColor[3] = rgba[3];
}

void tnsPolygonMode(GLenum PolyMode){
    T->StatePolyMode = PolyMode;
}
void tnsShadeMode(GLenum ShadeMode){
    T->StateShadeMode = ShadeMode;
}
void tnsVertex3d(real x, real y, real z){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    short vend = c->VertEnd;

    c->UseVert = 1;

    if (!c->Dimensions) c->Dimensions = 3;

    if(arrEnsureLength(&T->Vert, T->NextVert+c->Dimensions, &T->MaxVert, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxVert * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *varr = T->Vert;
    if (c->Dimensions == 3){
        varr[vend] = x;
        varr[vend + 1] = y;
        varr[vend + 2] = z;
        c->NumVert++;
        T->NextVert += 3;
        c->VertEnd += 3;
    }else{
        varr[vend] = x;
        varr[vend + 1] = y;
        c->NumVert++;
        T->NextVert += 2;
        c->VertEnd += 2;
    }
}
void tnsVertex2d(real x, real y){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    short vend = c->VertEnd;

    c->UseVert = 1;

    if (!c->Dimensions) c->Dimensions = 2;
    
    if(arrEnsureLength(&T->Vert, T->NextVert+c->Dimensions, &T->MaxVert, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxVert * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *varr = T->Vert;
    if (c->Dimensions == 2){
        varr[vend] = x;
        varr[vend + 1] = y;
        c->NumVert++;
        T->NextVert += 2;
        c->VertEnd += 2;
    }else{
        tnsVertex3d(x, y, 0.0f);
    }
}
void tnsVertexArray2d(real *verts, int amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    int trans = 2 * amount;
    short vend = c->VertEnd;

    c->UseVert = 1;

    if (!c->Dimensions) c->Dimensions = 2;

    if(arrEnsureLength(&T->Vert, T->NextVert+c->Dimensions*amount, &T->MaxVert, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxVert * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *varr = T->Vert;
    if (c->Dimensions == 2){
        int i;
        for (i = 0; i < trans; i++){
            varr[vend] = verts[i];
            vend++;
        }
        //memcpy(&varr[vend], verts, trans*sizeof(real));
        c->VertEnd += trans;
        c->NumVert += amount;
        T->NextVert += trans;
    }
}
void tnsVertexArray3d(real *verts, int amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    int trans = 3 * amount;
    short vend = c->VertEnd;

    c->UseVert = 1;

    if (!c->Dimensions) c->Dimensions = 3;

    if(arrEnsureLength(&T->Vert, T->NextVert+c->Dimensions*amount, &T->MaxVert, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxVert * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *varr = T->Vert;
    if (c->Dimensions == 3){
        int i;
        for (i = 0; i < trans; i++){
            varr[vend] = verts[i];
            vend++;
        }
        //memcpy(&varr[vend], verts, trans*sizeof(real));
        c->VertEnd += trans;
        c->NumVert += amount;
        T->NextVert += trans;
    }
}
void tnsColorArray4d(real *colors, int amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    int trans = 4 * amount;
    int ofst = c->ColorEnd;

    c->UseColor = 1;

    if(arrEnsureLength(&T->Color, T->NextColor+4*amount, &T->MaxColor, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->ColorBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxColor * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *carr = T->Color;
    int i;
    for (i = 0; i < trans; i++){
        carr[ofst] = colors[i];
        ofst++;
    }
    //memcpy(&T->Color[c->ColorEnd], colors, trans*sizeof(real));
    c->ColorEnd += trans;
    T->NextColor += trans;
}
void tnsNormalArray3d(real *normals, int amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    int trans = 3 * amount;
    int ofst = c->NormalEnd;

    c->UseNormal = 1;

    if(arrEnsureLength(&T->Normal, T->NextNormal+3*amount, &T->MaxNormal, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->NormalBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxNormal * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *narr = T->Normal;
    int i;
    for (i = 0; i < trans; i++){
        narr[ofst] = normals[i];
        ofst++;
    }
    //memcpy(&T->Normal[c->NormalEnd], normals, trans*sizeof(real));
    c->NormalEnd += trans;
    T->NextNormal += trans;
}
void tnsTexCoordArray2d(real *coords, int amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    int trans = 2 * amount;
    int ofst = c->TexCoordEnd;

    c->UseTexCoord = 1;
    c->UVDimensions = 2;

    if(arrEnsureLength(&T->TexCoord, T->NextTexCoord+2*amount, &T->MaxTexCoord, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->TexCoordBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxTexCoord * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *carr = T->TexCoord;
    int i;
    for (i = 0; i < trans; i++){
        carr[ofst] = coords[i];
        ofst++;
    }
    //memcpy(&T->TexCoord[c->TexCoordEnd], coords, trans*sizeof(real));
    c->TexCoordEnd += trans;
    T->NextTexCoord += trans;
}
void tnsTexCoordArray3d(real *coords, int amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    int trans = 3 * amount;
    int ofst = c->TexCoordEnd;

    c->UseTexCoord = 1;
    c->UVDimensions = 3;

    if(arrEnsureLength(&T->TexCoord, T->NextTexCoord+3*amount, &T->MaxTexCoord, sizeof(GLfloat))){
        glBindBuffer(GL_ARRAY_BUFFER, T->TexCoordBufObject);
        glBufferData(GL_ARRAY_BUFFER, T->MaxTexCoord * sizeof(GLfloat), 0, GL_DYNAMIC_DRAW);
    }

    GLfloat *carr = T->TexCoord;
    int i;
    for (i = 0; i < trans; i++){ carr[ofst] = coords[i]; ofst++; }
    c->TexCoordEnd += trans;
    T->NextTexCoord += trans;
}
void tnsIndexArray(GLuint *index, short amount){
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];

    //if (c->UseIndex) return;
    c->UseIndex = 1;

    if(arrEnsureLength(&T->Index, T->NextIndex+amount, &T->MaxIndex, sizeof(GLuint))){
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, T->IndexBufObject);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, T->MaxIndex * sizeof(GLuint), 0, GL_DYNAMIC_DRAW);
    }

    memcpy(&T->Index[c->IndexEnd], index, amount * sizeof(GLuint));
    c->IndexEnd += amount;
    c->NumIndex += amount;
    T->NextIndex += amount;
}

void tnsPackAs(GLenum Mode){
    arrEnsureLength(&T->DrawingCommand, T->NextCommand+1, &T->MaxCommand, sizeof(tnsCommand));
    tnsCommand *c = &T->DrawingCommand[T->NextCommand];
    tnsCommand *nc;

    T->NextCommand++;
    nc = &T->DrawingCommand[T->NextCommand];

    memcpy(c->UniformColor, T->StateColor, sizeof(GLfloat) * 4);

    if (Mode == GL_QUAD_STRIP || Mode == GL_QUADS) Mode=GL_TRIANGLE_STRIP;
    //if (Mode == GL_QUADS) Mode=GL_TRIANGLE_STRIP;
    c->Mode = Mode;
    c->ReplaceShader = T->StateShader;
    c->PolyMode = T->StatePolyMode;
    c->Shade = T->StateShadeMode;
    c->ColorTexture = T->StateTexColor;
    c->TextureMode = T->StateTextureMode;
    c->MultiplyColor = T->StateMultiplyColor;

    memset(nc, 0, sizeof(tnsCommand));
    nc->VertBegin = nc->VertEnd = c->VertEnd;
    nc->NormalBegin = nc->NormalEnd = c->NormalEnd;
    nc->ColorBegin = nc->ColorEnd = c->ColorEnd;
    nc->TexCoordBegin = nc->TexCoordEnd = c->TexCoordEnd;
    nc->IndexBegin = nc->IndexEnd = c->IndexEnd;
}
void tnsFlush(){
    tnsShader *cs = T->CurrentShader;
    tnsCommand *tc = &T->DrawingCommand[0];
    tnsCommand *c = tc;
    int PrevDimensions = 0;
    int LastVertBegin = 0;

    if (!c || !cs) return;

    if (T->NextVert){
        glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
        //glEnableVertexAttribArray(cs->iVertex);
        glBufferSubData(GL_ARRAY_BUFFER, 0, T->NextVert * sizeof(GLfloat), T->Vert);
    }
    if (T->NextColor){
        glBindBuffer(GL_ARRAY_BUFFER, T->ColorBufObject);
        //glEnableVertexAttribArray(cs->iColor);
        glBufferSubData(GL_ARRAY_BUFFER, 0, T->NextColor * sizeof(GLfloat), T->Color);
    }
    if (T->NextNormal){
        glBindBuffer(GL_ARRAY_BUFFER, T->NormalBufObject);
        //glEnableVertexAttribArray(cs->iColor);
        glBufferSubData(GL_ARRAY_BUFFER, 0, T->NextNormal * sizeof(GLfloat), T->Normal);
    }
    if (T->NextTexCoord){
        glBindBuffer(GL_ARRAY_BUFFER, T->TexCoordBufObject);
        //glEnableVertexAttribArray(cs->iColor);
        glBufferSubData(GL_ARRAY_BUFFER, 0, T->NextTexCoord * sizeof(GLfloat), T->TexCoord);
    }

    for (int i=0;i<T->NextCommand;i++){
        c=&T->DrawingCommand[i];

        if (c->PolyMode != T->StatePolyMode){glPolygonMode(GL_FRONT_AND_BACK, c->PolyMode); T->StatePolyMode=c->PolyMode;}
        if (c->Shade != T->StateShadeMode){glShadeModel(c->Shade); T->StateShadeMode=c->Shade;}
        if (c->ReplaceShader && c->ReplaceShader != T->CurrentShader){
            tnsEnableShaderv(c->ReplaceShader);
            cs = c->ReplaceShader;
            if (!cs) continue;
        }

        glBindBuffer(GL_ARRAY_BUFFER, T->VertBufObject);
        if (c->UseVert){
            glEnableVertexAttribArray(cs->iVertex);
            glVertexAttribPointer(cs->iVertex, (c->Dimensions ? c->Dimensions : PrevDimensions),
                                  GL_FLOAT, 0, 0, c->VertBegin * sizeof(GLfloat));
            LastVertBegin = c->VertBegin;
        }else{
            glEnableVertexAttribArray(cs->iVertex);
            glVertexAttribPointer(cs->iVertex, (c->Dimensions ? c->Dimensions : PrevDimensions),
                                  GL_FLOAT, 0, 0, LastVertBegin * sizeof(GLfloat));
        }

        PrevDimensions = (c->Dimensions ? c->Dimensions : PrevDimensions);

        if (c->UseIndex){
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, T->IndexBufObject);
            glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, c->NumIndex * sizeof(GLuint), &T->Index[c->IndexBegin]);
        }

        if (cs->iNormal != -1){
            glBindBuffer(GL_ARRAY_BUFFER, T->NormalBufObject);
            if (c->UseNormal){
                glEnableVertexAttribArray(cs->iNormal);
            }
            glVertexAttribPointer(cs->iNormal, 3, GL_FLOAT, 0, 0, c->NormalBegin * sizeof(GLfloat));
            if (!c->UseNormal && c->Dimensions == 2){
                glDisableVertexAttribArray(cs->iNormal);
                glVertexAttrib3f(cs->iNormal, 0, 0, -1);
            }
        }

        if (cs->iColor != -1){
            glBindBuffer(GL_ARRAY_BUFFER, T->ColorBufObject);
            if (c->UseColor){
                glEnableVertexAttribArray(cs->iColor);
                glVertexAttribPointer(cs->iColor, 4, GL_FLOAT, 0, 0, c->ColorBegin * sizeof(GLfloat));
            }else{
                glDisableVertexAttribArray(cs->iColor);
                glVertexAttrib4fv(cs->iColor, c->UniformColor);
            }
        }

        if (cs->iUV != -1){
            if (c->UseTexCoord){
                glBindBuffer(GL_ARRAY_BUFFER, T->TexCoordBufObject);
                if (c->UseTexCoord){ glEnableVertexAttribArray(cs->iUV);
                    glVertexAttribPointer(cs->iUV, c->UVDimensions, GL_FLOAT, 0, 0, c->TexCoordBegin * sizeof(GLfloat)); }
                else{ glDisableVertexAttribArray(cs->iUV); }
            }
        }

        if (c->TextureMode && c->ColorTexture && cs->iTexColor != -1){
            tnsBindTexture(c->ColorTexture); tnsUniformUseTexture(cs, c->TextureMode, c->ColorTexture->Multisample);
        }else{
            tnsUniformUseTexture(cs, 0, 0); //tnsUnbindTexture(); 
        }

        if(cs->iMultiplyColor != -1){
            tnsUniformUseMultiplyColor(cs, c->MultiplyColor);
        }

        if (c->UseIndex){
            glDrawElements(c->Mode, c->NumIndex, GL_UNSIGNED_INT, 0);
        }else{
            glDrawArrays(c->Mode, 0, c->NumVert);
        }
    }

    T->NextCommand=0;
    c = &T->DrawingCommand[0];

    memset(c, 0, sizeof(tnsCommand));

    c->ColorBegin = c->ColorEnd = T->NextColor = 0;
    c->NormalBegin = c->NormalEnd = T->NextNormal = 0;
    c->TexCoordBegin = c->TexCoordEnd = T->NextTexCoord = 0;
    c->VertBegin = c->VertEnd = T->NextVert = 0;
    c->IndexBegin = c->IndexEnd = T->NextIndex = 0;

    //must --why?
    //T->BindedShader = 0;
};

//============================================================================================[offscr]

const GLuint TNS_ATTACHMENT_ARRAY_NONE[] = {GL_NONE};
const GLuint TNS_ATTACHMENT_ARRAY[] = {GL_COLOR_ATTACHMENT0};
const GLuint TNS_ATTACHMENT_ARRAY_1[] = {GL_COLOR_ATTACHMENT1};
const GLuint TNS_ATTACHMENT_ARRAY_2[] = {GL_COLOR_ATTACHMENT2};
const GLuint TNS_ATTACHMENT_ARRAY_1_2[] = {GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2};
const GLuint TNS_ATTACHMENT_ARRAY_0_1_2[] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2};
const GLenum TNS_WINDOW_DRAWBUFFER_ARRAY[] = {GL_BACK};

tnsOffscreen *tnsCreateOffscreenHandle(){
    tnsOffscreen *toff = CreateNew(tnsOffscreen);
    glGenFramebuffers(1, &toff->FboHandle);
    lstAppendItem(&T->Offscreens, toff);
    return toff;
}
void tnsAttach2DOffscreenBuffer(tnsOffscreen *target, GLuint attatchment, tnsTexture *use){
    if (!target || !use || target->FboHandle == -1 || use->GLTexHandle == -1) return;

    if (attatchment >= GL_COLOR_ATTACHMENT0 && attatchment <= GL_COLOR_ATTACHMENT15){
        //if (target->pColor[attatchment - GL_COLOR_ATTACHMENT0]) return;

        glBindFramebuffer(GL_FRAMEBUFFER, target->FboHandle);
        tnsBindTexture(use);
        glFramebufferTexture2D(GL_FRAMEBUFFER, attatchment, use->GLTexType, use->GLTexHandle, 0);
        target->pColor[attatchment - GL_COLOR_ATTACHMENT0] = use;
        tnsUnbindTexture();
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }elif (attatchment == GL_DEPTH_ATTACHMENT){
        //if (target->pDepth) return;
        glBindFramebuffer(GL_FRAMEBUFFER, target->FboHandle);
        tnsBindTexture(use);
        glFramebufferTexture2D(GL_FRAMEBUFFER, attatchment, use->GLTexType, use->GLTexHandle, 0);
        //glBindRenderbufferEXT(GL_RENDERBUFFER, use->GLTexHandle);
        //glBindFramebuffer(GL_FRAMEBUFFER, target->FboHandle);
        //glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, use->GLTexHandle);
        target->pDepth = use;
        tnsUnbindTexture();
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
        //glBindRenderbufferEXT(GL_RENDERBUFFER, 0);
        //glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }
    GLenum result = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    //if (result == GL_FRAMEBUFFER_COMPLETE) {
    //	printf("Framebuffer complete!\n");
    //}
    //else {
    //	printf("Framebuffer incomplete!\n");
    //}
}
void tnsDetach2DOffscreenBuffer(tnsOffscreen *target, GLuint which_attach_point){
    if (which_attach_point >= GL_COLOR_ATTACHMENT0 && which_attach_point <= GL_COLOR_ATTACHMENT15){
        if (target->pColor[which_attach_point - GL_COLOR_ATTACHMENT0] == 0) return;

        glBindFramebuffer(GL_FRAMEBUFFER, target->FboHandle);
        glFramebufferTexture2D(GL_FRAMEBUFFER, which_attach_point, GL_TEXTURE_2D, 0, 0);

        tnsDeleteTexture(target->pColor[which_attach_point - GL_COLOR_ATTACHMENT0]);

        target->pColor[which_attach_point - GL_COLOR_ATTACHMENT0] = 0;
    }elif (which_attach_point == GL_DEPTH_ATTACHMENT){
        if (target->pDepth) return;

        glBindFramebuffer(GL_FRAMEBUFFER, target->FboHandle);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, 0, 0);

        tnsDeleteTexture(target->pDepth);

        target->pDepth = 0;
    }
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    tnsUnbindTexture();
}

tnsOffscreen *tnsCreate2DOffscreen(int glInternalFormat, int w, int h, int Multisample, int WithDepth){
    tnsOffscreen *toff = tnsCreateOffscreenHandle();
    tnsTexture *color;
    tnsTexture *depth;

    if(glInternalFormat){
        color = tnsCreate2DTexture(glInternalFormat, w, h, Multisample);
        tnsAttach2DOffscreenBuffer(toff, GL_COLOR_ATTACHMENT0, color);
    }
    if(WithDepth){
        depth = tnsCreate2DTexture(GL_DEPTH_COMPONENT, w, h, Multisample);
        tnsAttach2DOffscreenBuffer(toff, GL_DEPTH_ATTACHMENT, depth);
    }

    return toff;
}

void tnsConfigureOffscreen(tnsOffscreen *off, int w, int h){
    tnsTexture* t=off->pColor[0];
    t->Width = w; t->Height = h; tnsConfigure2DTexture(t);
    t=off->pDepth;
    if(t){
        t->Width = w; t->Height = h; tnsConfigure2DTexture(t);
    }
}

void tnsDelete2DOffscreen(tnsOffscreen *o){
    if (!o) return;

    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

    if (o->pColor[1]){
        tnsDetach2DOffscreenBuffer(o, GL_COLOR_ATTACHMENT1);
        tnsDeleteTexture(o->pColor[1]);
    }
    if (o->pColor[2]){
        tnsDetach2DOffscreenBuffer(o, GL_COLOR_ATTACHMENT2);
        tnsDeleteTexture(o->pColor[2]);
    }
    if (o->pColor[3]){
        tnsDetach2DOffscreenBuffer(o, GL_COLOR_ATTACHMENT3);
        tnsDeleteTexture(o->pColor[3]);
    }

    tnsDetach2DOffscreenBuffer(o, GL_COLOR_ATTACHMENT0);
    tnsDetach2DOffscreenBuffer(o, GL_DEPTH_ATTACHMENT);

    tnsDeleteTexture(o->pColor[0]);
    tnsDeleteTexture(o->pDepth);

    glDeleteFramebuffers(1, &o->FboHandle);
    //tnsDeleteTexture(o->pStencil);

    lstRemoveItem(&T->Offscreens, o);
    FreeMem(o);
}

void tnsDrawToOffscreen(tnsOffscreen *toff, int HowMany, GLuint *AttachmentArray){
    if (!toff) return;
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, toff->FboHandle);
    if(AttachmentArray==TNS_ATTACHMENT_ARRAY_NONE){glDrawBuffer(GL_NONE);}
    else glDrawBuffers((HowMany ? HowMany : 1), (AttachmentArray ? AttachmentArray : TNS_ATTACHMENT_ARRAY));
    T->IsOffscreen = 1;
    T->BindedShader = 0;
}
//void tnsDrawToExtraColorAttachment(tnsOffscreen *toff){
//    if (!toff) return;
//    if (!toff->pColor[1]){
//        tnsCreate2DOffscreenMSAttachmentExtraColor(toff);
//    }
//    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, toff->FboHandle);
//    glDrawBuffers(1, TNS_ATTACHMENT_ARRAY_1);
//    T->IsOffscreen = 1;
//    T->BindedShader = 0;
//}
//void tnsDrawToExtraNormalAttachment(tnsOffscreen *toff){
//    if (!toff) return;
//    if (!toff->pColor[2]){
//        tnsCreate2DOffscreenMSAttachmentExtraNormal(toff);
//    }
//    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, toff->FboHandle);
//    glDrawBuffers(1, TNS_ATTACHMENT_ARRAY_2);
//    T->IsOffscreen = 1;
//    T->BindedShader = 0;
//}
//void tnsDrawToAllExtraAttachments(tnsOffscreen *toff){
//    if (!toff) return;
//    if (!toff->pColor[1]){
//        tnsCreate2DOffscreenMSAttachmentExtraColor(toff);
//    }
//    if (!toff->pColor[2]){
//        tnsCreate2DOffscreenMSAttachmentExtraNormal(toff);
//    }
//    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, toff->FboHandle);
//    glDrawBuffers(2, TNS_ATTACHMENT_ARRAY_1_2);
//    T->IsOffscreen = 1;
//    T->BindedShader = 0;
//}
void tnsDrawToOffscreenOnlyBind(tnsOffscreen *toff, int HowMany, GLuint *AttachmentArray){
    if (!toff) return;
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, toff->FboHandle);
}
void tnsReadFromOffscreen(tnsOffscreen *toff){
    if (!toff) return;
    glBindFramebuffer(GL_READ_FRAMEBUFFER, toff->FboHandle);
}
void tnsPassColorBetweenOffscreens(tnsOffscreen *from, tnsOffscreen *to,
                                   GLint srcX0, GLint srcY0, GLint srcX1, GLint srcY1, GLint dstX0, GLint dstY0, GLint dstX1, GLint dstY1, GLenum FilterMode){
    if (!from || !to) return;
    glBlitFramebufferEXT(
        srcX0, srcY0, srcX1, srcY1,
        dstX0, dstY0, dstX1, dstY1,
        GL_COLOR_BUFFER_BIT, FilterMode);
}
void tnsDrawToScreen(){
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
    glDrawBuffer(GL_BACK);
    T->IsOffscreen = 0;
    T->BindedShader = 0;
}

//===========================================================================[FONT]

tnsFontManager *FM;

void tnsSetuptnsFontManager(){
    FM = CreateNew(tnsFontManager);
};
void tnsQuitFontManager(){
    tnsInvalidateFontCache();
    tnsFont*f=FM->UsingFont;
    for(int i=0;i<f->NumFaces;i++){
        FT_Done_Face(f->ftface[i]);
    }
    if(f->ftfacemono) FT_Done_Face(f->ftfacemono);
    FT_Done_FreeType(f->ftlib);
    free(f->characters);
    free(f->monocharacters);
    free(f);
    FreeMem(FM);
}

int next_p2(int a){
    int rval = 1;
    while (rval < a)
        rval <<= 1;
    return rval;
}
int tnsGetMonoFontAdvance(){
    return FM->UsingFont->MonoAdvance;
}
int tnsInvalidateFontCache(){
    tnsFont*f=FM->UsingFont;
    for(int i=0;i<TNS_UNICODE_COUNT;i++){ if(f->characters){ free(f->characters[i]); f->characters[i]=0; } }
    for(int i=0;i<TNS_MONO_COUNT;i++){ if(f->monocharacters){ free(f->monocharacters[i]); f->monocharacters[i]=0; } }
    f->CurrentX=f->CurrentY=0;
    f->height = LA_RH*(MAIN.FontSize/2.0f+0.5f);
    int GenHeight=LA_RH*MAIN.FontSize;
    for(int i=0;i<f->NumFaces;i++){
        FT_Set_Char_Size(f->ftface[i], 0, GenHeight << 6, 96, 96);
    }
    if(f->ftfacemono){
        FT_Set_Char_Size(f->ftfacemono, 0, (GenHeight << 6)*f->MonoScale, 96, 96); FT_Glyph glyph;
        if (FT_Load_Char(f->ftfacemono, 'a', FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP)||
            FT_Get_Glyph(f->ftfacemono->glyph, &glyph)){ SEND_PANIC_ERROR("Monospace font doesn't contain character 'a'"); }
        f->MonoAdvance=f->ftfacemono->glyph->advance.x/64.0;
        if(glyph) FT_Done_Glyph(glyph);
    }
}
tnsFontSingleCharacter *tfntFetchCharTextureIDW(uint32_t ch, int UseMono);
void tnsLoadSystemFontMono(char* mono){
    if(!FM->UsingFont || !mono || !mono[0]) return;
    tnsFont *f=FM->UsingFont;
    
    int GenHeight=LA_RH*MAIN.FontSize;

    int full_adv=0,half_adv=0;
    for(int i=0;i<f->NumFaces;i++){
        if(!FT_Get_Advance(f->ftface[i],L'我',FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP, &full_adv)) break;
    }

    if (FT_New_Face(f->ftlib, mono, 0, &f->ftfacemono)) SEND_PANIC_ERROR("Freetype Can't Init Face");
    FT_Select_Charmap(f->ftfacemono, FT_ENCODING_UNICODE);
    FT_Set_Char_Size(f->ftfacemono, 0, GenHeight << 6, 96, 96);
    if(!FT_Get_Advance(f->ftfacemono,'a',FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP, &half_adv)){
        f->MonoScale=(real)full_adv/(half_adv*2);
        FT_Set_Char_Size(f->ftfacemono, 0, (GenHeight << 6)*f->MonoScale, 96, 96);
        logPrint("Monospace font scale: %.2f\n",f->MonoScale); FT_Glyph glyph;
        if (FT_Load_Char(f->ftfacemono, 'a', FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP)||
            FT_Get_Glyph(f->ftfacemono->glyph, &glyph)){ SEND_PANIC_ERROR("Monospace font doesn't contain character 'a'"); }
        f->MonoAdvance=f->ftfacemono->glyph->advance.x/64.0;
        if(glyph) FT_Done_Glyph(glyph);
    }
};
void tnsLoadSystemFont(int num_fonts, ...){
    char * name;

    tnsFont *f = CreateNew(tnsFont);
    f->characters=calloc(1, sizeof(tnsFontSingleCharacter*)*TNS_UNICODE_COUNT);
    f->monocharacters=calloc(1, sizeof(tnsFontSingleCharacter*)*TNS_MONO_COUNT);

    f->height = LA_RH*(MAIN.FontSize/2.0f+0.5f);
    int GenHeight=LA_RH*MAIN.FontSize;

    if (FT_Init_FreeType(&f->ftlib)) SEND_PANIC_ERROR("Freetype Init Failed!");

    if(num_fonts>=16){num_fonts=16;}
    va_list fn;
    va_start(fn, num_fonts);
    for(int i=0;i<num_fonts;i++){
        name = va_arg(fn, char*);
        if (name && FT_New_Face(f->ftlib, name, 0, &f->ftface[i])) SEND_PANIC_ERROR("Freetype Can't Init Face");
        FT_Select_Charmap(f->ftface[i], FT_ENCODING_UNICODE);
        FT_Set_Char_Size(f->ftface[i], 0, GenHeight << 6, 96, 96);
    }
    va_end(fn);
    f->NumFaces = num_fonts;
    
    tnsInit2DTexture(&f->TexBuffer, GL_RED, TNS_FONT_BUFFER_W, TNS_FONT_BUFFER_H, 0);
    
    lstAppendItem(&FM->Fonts, f);
    tnsUseFont(0);
};
int tnsLoadVectorGraphPackage(const char *name, unsigned int size){
    char i;

    tnsFont *f = CreateNew(tnsFont);

    f->height = (int)((real)size * 1.4);
    f->fontName = name;

    if (FT_Init_FreeType(&f->ftlib)) SEND_PANIC_ERROR("Can't Load Main Font:Freetype Init Failed!");
    if (FT_New_Face(f->ftlib, name, 0, &f->ftface)) SEND_PANIC_ERROR("Can't Load Main Font:Freetype Can't Init Face");

    FT_Select_Charmap(f->ftface, FT_ENCODING_UNICODE);
    FT_Set_Char_Size(f->ftface, size << 6, size << 6, 96, 96);

    //tnsInit2DTexture(&f->TexBuffer, GL_ALPHA, TNS_FONT_BUFFER_W, TNS_FONT_BUFFER_H, 0);

    //lstAppendItem(&FM->Fonts, f);

    //tnsUseFont(name);

    FM->VectorsGrapghs = f;
};

int tfntBufferWidthEnough(int total_width, int current_width, int this_width){
    return (current_width + this_width < total_width);
}

void tfntApplyCharacterBufferOffset(tnsFont *f, tnsFontSingleCharacter *fsc){
    if (!tfntBufferWidthEnough(TNS_FONT_BUFFER_W, f->CurrentX, fsc->width)){
        f->CurrentY += f->height+1;
        f->CurrentX = 0;
    }
    fsc->bufferx = f->CurrentX;
    fsc->buffery = f->CurrentY;
    f->CurrentX += fsc->width;
}

tnsFontSingleCharacter *tfntFetchVectorGraphTextureIDW(uint32_t ID){
    //GLuint revel = 0;
    //tnsFont *f = FM->VectorsGrapghs;
    //tnsFontSingleCharacter *fsc = 0;
    //FT_Glyph glyph = 0;
    //FT_BitmapGlyph bitmap_glyph;
    //FT_Bitmap bm;
    //FT_Face face;
    //int w, h, i, j;
    //GLubyte *buf = 0;
    //int a;
    //int Size = ((int)(_ICON_SYMBOL_SIZE[ID - 0)) << 6;
    //int ret;

    //if (!f)
    //    return 0;

    //if (revel = f->icons[ID].Generated)
    //    return &f->icons[ID];

    //FT_Set_Char_Size(f->ftface, Size, Size, 96, 96);

    //if (ret = FT_Load_Char(f->ftface, ID, FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP))
    //    return 0;

    //fsc = &f->icons[ID];

    //face = f->ftface;

    //if (FT_Get_Glyph(face->glyph, &glyph))
    //    return 0;

    //FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);

    //FT_Glyph_To_Bitmap(&glyph, FT_RENDER_MODE_NORMAL, 0, 1);
    //bitmap_glyph = glyph;
    //bm = bitmap_glyph->bitmap;

    //w = bm.width;
    //h = bm.rows;
    //fsc->width = w;
    //fsc->height = h;
    //fsc->advx = face->glyph->advance.x / 64.0f;
    //fsc->advy = face->size->metrics.y_ppem;
    //fsc->deltax = bitmap_glyph->left;
    //fsc->deltay = bitmap_glyph->top - h;

    ////tfntApplyCharacterBufferOffset(f, fsc);

    ////glBindTexture(GL_TEXTURE_2D, f->TexBuffer.GLTexHandle);

    //buf = CreateNewBuffer(GLubyte, w * h);

    //for (j = 0; j < h; j++)
    //{
    //    for (i = 0; i < w; i++)
    //    {
    //        char _vl = bm.buffer[i + w * j];
    //        buf[i + (h - j - 1) * w] = _vl;
    //    }
    //}

    //glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    //fsc->Tex = tnsCreate2DTexture(GL_ALPHA, w, h, buf);

    //glBindTexture(GL_TEXTURE_2D, 0);

    ////glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, w, h, 0, GL_ALPHA, GL_UNSIGNED_BYTE, buf);
    //glBindTexture(GL_TEXTURE_2D, 0);

    //FreeMem(buf);

    //fsc->Generated = 1;
    //return fsc;
}
tnsFontSingleCharacter *tfntFetchCharTextureIDW(uint32_t ch, int UseMono){
    GLuint revel = 0;
    tnsFont *f = FM->UsingFont;
    tnsFontSingleCharacter **use_fsc=UseMono?f->monocharacters:f->characters, *fsc = 0;
    FT_Glyph glyph = 0;
    FT_BitmapGlyph bitmap_glyph;
    FT_Bitmap bm;
    FT_Face face;
    int w, h, i, j;
    GLubyte *buf = 0;

    if (!f) return 0;

    if(ch>TNS_MONO_COUNT){ UseMono=0; use_fsc=f->characters; }
    if(ch>TNS_UNICODE_COUNT){ return 0; }

    fsc = use_fsc[ch];
    if(!fsc){ fsc=use_fsc[ch]=calloc(1,sizeof(tnsFontSingleCharacter)); }

    if (revel = fsc->Generated) return fsc;

    if(UseMono && f->ftfacemono){
        face = f->ftfacemono;
        if(!FT_Get_Char_Index(face,ch)){return 0;}
        if (FT_Load_Char(face, ch, FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP)){ return 0; }
        if (FT_Get_Glyph(face->glyph, &glyph)){ return 0; }
    }else{
        for(int i=0;i<f->NumFaces;i++){
            if(!f->ftface[i]){continue;}
            face = f->ftface[i];
            if(!FT_Get_Char_Index(face,ch)){continue;}
            if (FT_Load_Char(face, ch, FT_LOAD_TARGET_NORMAL | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP)){ continue; }
            if (FT_Get_Glyph(face->glyph, &glyph)){ continue; }
            break;
        }
        if(!glyph){ return 0; }
    }

    FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);

    FT_Glyph_To_Bitmap(&glyph, FT_RENDER_MODE_NORMAL, 0, 1);
    bitmap_glyph = glyph;
    bm = bitmap_glyph->bitmap;

    w = bm.width;
    h = bm.rows;
    fsc->width = w;
    fsc->height = h;
    fsc->advx = face->glyph->advance.x / 64.0f;
    fsc->advy = face->size->metrics.y_ppem;
    fsc->deltax = bitmap_glyph->left;
    fsc->deltay = bitmap_glyph->top - h;

    tfntApplyCharacterBufferOffset(f, fsc);

    tnsBindTexture(&f->TexBuffer);
    //glBindTexture(GL_TEXTURE_2D, f->TexBuffer.GLTexHandle);

    buf = CreateNewBuffer(GLubyte, w * h);

    for (j = 0; j < h; j++){
        for (i = 0; i < w; i++){
            char _vl = bm.buffer[i + w * j];
            buf[i + (h - j - 1) * w] = _vl;
        }
    }

    //glPixelStorei(GL_UNPACK_LSB_FIRST, GL_FALSE);
    //glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexSubImage2D(GL_TEXTURE_2D, 0, fsc->bufferx, fsc->buffery, w, h, GL_RED, GL_UNSIGNED_BYTE, buf);

    FreeMem(buf);
    FT_Done_Glyph(glyph);

    fsc->Generated = 1;
    return fsc;
}
tnsFontSingleCharacter *tfntFetchCharacterW(uint32_t ch, int UseMono){
    tnsFontSingleCharacter* fsc = tfntFetchCharTextureIDW(ch, UseMono);
    if(fsc) return fsc;
    return tfntFetchCharTextureIDW('?', UseMono);
}

int CMP_NAME_IsThisFont(tnsFont *enumed, char *name){
    return (!strcmp(enumed->fontName, name));
};

int tnsUseFont(char *name){
    tnsFont *f;
    if(!name){f = FM->Fonts.pFirst;}else{ f = lstFindItem(name, CMP_NAME_IsThisFont, &FM->Fonts); }

    if (!f) return 0;

    FM->UsingFont = f;

    return 1;
};

int tnsStringGetDimension(char* content, uint32_t* contentU, int Count, int WLimit, int* Rows, int UseMono){
    int sx = 0, sy = FM->UsingFont->height, MA=FM->UsingFont->MonoAdvance;
    int i, rows=1, advance=1;
    int C = 0;
    int len = contentU?strlenU(contentU):(content?strlen(content):0);
    char hb;
    int RestoreI;
    int MaxSX=0;

    if(!WLimit){WLimit=INT_MAX;}

    int UC=1;
    for (i = 0; i < len && UC; i+=advance){
        UC = contentU?contentU[i]:laToUnicode(&content[i], &advance);
        tnsFontSingleCharacter *fsc;
        if (UC == L'\n'){
            if(sx>MaxSX) MaxSX=sx;
            sx = 0; sy += LA_RH; rows++;
            continue;
        }else{
            fsc = tfntFetchCharacterW(UC, UseMono);
            int dx=fsc->advx; if(UseMono){ dx/=MA; if(dx<1) dx=1; if(dx>1)dx=2; dx*=MA; }
            if(sx+dx > WLimit){ sx = 0; sy += LA_RH; rows++; }
            sx += dx;
            C += 1;
            if (Count && C == Count) break;
        }
    }
    if(sx==0 && rows>1){rows--;}
    if(sx>MaxSX) MaxSX=sx;
    if(Rows) (*Rows)=rows;
    return MaxSX;
}
int tnsStringGetWidth(char *content, int Count, int UseMono){
    return tnsStringGetDimension(content, 0, Count, 0, 0, UseMono);
}
int tnsStringGetWidthU(uint32_t *contentU, int Count, int UseMono){
    return tnsStringGetDimension(0, contentU, Count, 0, 0, UseMono);
}
void tnsDrawStringM(char *content, uint32_t* contentU, real Color[4], int L, int R, int T, int Flags){
    int sx = L, sy = (LA_RH!=LA_RH0)?(((((real)FM->UsingFont->height/LA_RH0-0.5)/MAIN.UiScale)+0.5)*LA_RH + T):(FM->UsingFont->height+T);
    int MA=FM->UsingFont->MonoAdvance;
    int i,advance=1;
    int len = contentU?strlenU(contentU):strlen(content);
    real xo, yo, xl, yl;
    real TexCoord[12];
    real VertArr[12];
    int total_width = 0;
    tnsFont *f = FM->UsingFont;
    int FscHeight, RestoreI;
    int RevY=(Flags&LA_TEXT_REVERT_Y);
    int UseMono=(Flags&LA_TEXT_MONO);

    int any=0, UC=1; int BreakNow=0;
    for (i = 0; i < len && UC; i+=advance){
        UC = contentU?contentU[i]:laToUnicode(&content[i], &advance);
        tnsFontSingleCharacter *fsc;
        int cx, cy;

        if (UC == L'\n'){ sx = L; sy += LA_RH; continue; }

        fsc = tfntFetchCharacterW(UC, Flags&LA_TEXT_MONO);

        int dx=fsc->advx; if(UseMono){ dx/=MA; if(dx<1) dx=1; if(dx>1)dx=2; dx*=MA; }
        if (sx + dx > R){
            if(Flags&LA_TEXT_LINE_WRAP){
                sx=L; sy+=LA_RH;
            }else{
                if(Flags&LA_TEXT_OVERFLOW_ARROW){ fsc = tfntFetchCharacterW(L'▷', 0); sx=R-fsc->advx; BreakNow=1; }
                else break;
            }
        }

        cx = sx + fsc->deltax;
        cy = sy - fsc->deltay;
        xo = (real)fsc->bufferx / (real)TNS_FONT_BUFFER_W;
        yo = (real)fsc->buffery / (real)TNS_FONT_BUFFER_H;
        xl = (real)(fsc->bufferx + fsc->width) / (real)TNS_FONT_BUFFER_W;
        yl = (real)(fsc->buffery + fsc->height) / (real)TNS_FONT_BUFFER_H;

        tnsMakeQuadT2d(TexCoord, xo, yl, xo, yo, xl, yl, xl, yo);
        if (RevY) tnsMakeQuadT2d(VertArr,
                          cx, 2 * T - cy + fsc->height,
                          cx, 2 * T - cy,
                          cx + fsc->width, 2 * T - cy + fsc->height,
                          cx + fsc->width, 2 * T - cy);
        else tnsMakeQuadT2d(VertArr,
                          cx, +cy - fsc->height,
                          cx, +cy,
                          cx + fsc->width, +cy - fsc->height,
                          cx + fsc->width, +cy);

        tnsUseMaskTexture(&f->TexBuffer);
        tnsColor4dv(Color);
        tnsVertexArray2d(VertArr, 6);
        tnsTexCoordArray2d(TexCoord, 6);
        any=1;

        sx += dx;
        
        if(BreakNow){ break; }
    }
    if(any) tnsPackAs(GL_TRIANGLES);
}
void tnsDrawStringAutoM(char *content, uint32_t* contentU, real Color[4], int L, int R, int T, int Flags){
    int LL;
    int al = Flags&LA_TEXT_ALIGN;
    
    if (!content && !contentU) return;
    switch (al){
    case LA_TEXT_ALIGN_AUTO:
    case LA_TEXT_ALIGN_CENTER: LL = L+(R-L-tnsStringGetDimension(content, contentU, 0, 0, 0, Flags&LA_TEXT_MONO)) / 2; if(LL<L) LL=L; break;
    default:
    case LA_TEXT_ALIGN_LEFT: LL = L; break;
    case LA_TEXT_ALIGN_RIGHT: LL = R-(tnsStringGetDimension(content, contentU, 0, 0, 0, Flags&LA_TEXT_MONO)); break;
    }
    tnsDrawStringM(content, contentU, Color, LL, R, T, Flags);
}
void tnsDrawStringAuto(char *content, real Color[4], int L, int R, int T, int Flags){
    tnsDrawStringAutoM(content,0,Color,L,R,T,Flags);
}
void tnsDrawStringWithPriority(char *Label, char *MajorContent, real Color[4], int TextAlign, int L, int R, int T, int Flags){
    int W = R - L;
    int Str1W = tnsStringGetWidth(Label, 0, Flags&LA_TEXT_MONO);
    int Str2W = tnsStringGetWidth(MajorContent, 0, Flags&LA_TEXT_MONO);
    if (Str1W + Str2W > W){
        if (Str2W < W){
            tnsDrawStringM(MajorContent, 0, Color, R - Str2W, R, T, Flags);
            tnsDrawStringM(Label, 0, Color, L, R - Str2W, T, Flags);
        }else
            tnsDrawStringM(MajorContent, 0, Color, L, R, T, Flags);
    }else{
        int LL = L, ML;
        switch (TextAlign){
        case LA_TEXT_ALIGN_CENTER:
            ML = L + Str1W + (W - (Str1W + Str2W)) / 2;
            break;
        case LA_TEXT_ALIGN_LEFT:
            ML = L + Str1W;
            break;
        default:
        case LA_TEXT_ALIGN_AUTO:
        case LA_TEXT_ALIGN_RIGHT:
            ML = R - Str2W;
            break;
        }

        tnsDrawStringM(Label, 0, Color, LL, R, T, Flags);
        tnsDrawStringM(MajorContent, 0, Color, ML, R, T, Flags);
    }
}
void tnsDrawIcon(uint32_t ID, real Color[4], int L,int R, int T, int Flags){
    char buf[5]={0}; char* next; laToUTF8(ID,buf,&next);
    tnsDrawStringAuto(buf, Color, L,R,T,Flags);
}

//=====================================================================================[Object]

tnsObject *tnsFindObject(char *Name, tnsObject *FromObj){
    tnsObject *io; tnsObject *ro;
    laListHandle* l=FromObj?(&FromObj->ChildObjects):(&T->World.AllObjects);
    for (laListItemPointer*lip=l->pFirst;lip;lip=lip->pNext){
        io=((l==&T->World.AllObjects)?lip:lip->p); if (strSame(io->Name->Ptr, Name)){ return io; }
        if (ro = tnsFindObject(Name, io)) return ro;
    }
    return 0;
}

void tnsExtractSelfTransformValue(tnsObject *o){
    if (!o) return;
    tnsExtractLocation44d(o->SelfTransform, &o->Location[0], &o->Location[1], &o->Location[2]);
    tnsExtractXYZEuler44d(o->SelfTransform, &o->Rotation[0], &o->Rotation[1], &o->Rotation[2]);
    tnsExtractUniformScale44d(o->SelfTransform, &o->Scale);
}
void tnsExtractGlobalTransformValue(tnsObject *o){
    if (!o) return;
    tnsExtractLocation44d(o->GlobalTransform, &o->GLocation[0], &o->GLocation[1], &o->GLocation[2]);
    tnsExtractXYZEuler44d(o->GlobalTransform, &o->GRotation[0], &o->GRotation[1], &o->GRotation[2]);
    tnsExtractUniformScale44d(o->GlobalTransform, &o->GScale);
}
void tnsCopyGlobalTransform(tnsObject *to, tnsObject *from){
    memcpy(to->GlobalTransform, from->GlobalTransform, sizeof(tnsMatrix44d));
    memcpy(to->GLocation, from->GLocation, sizeof(tnsVector3d));
    memcpy(to->GRotation, from->GRotation, sizeof(tnsVector3d));
    to->GScale = from->GScale;
    tnsGlobalTransformValueChanged(to);
}

void tnsSelfMatrixChanged(tnsObject* o, int ApplyToChild){
    if (!o->ParentObject){ memcpy(o->GlobalTransform, o->SelfTransform, sizeof(tnsMatrix44d)); }
    else{ tnsMultiply44d(o->GlobalTransform, o->ParentObject->GlobalTransform, o->SelfTransform); }
    tnsExtractSelfTransformValue(o);
    tnsExtractGlobalTransformValue(o);
    if(ApplyToChild) for (laListItemPointer* li=o->ChildObjects.pFirst;li;li=li->pNext){ tnsSelfMatrixChanged(li->p,1); }
}
void tnsGlobalMatrixChanged(tnsObject* o, int ApplyToChild){
    if (!o->ParentObject){ memcpy(o->SelfTransform, o->GlobalTransform, sizeof(tnsMatrix44d)); }
    else{ tnsMatrix44d inv;
        tnsInverse44d(inv, o->ParentObject->GlobalTransform);
        tnsMultiply44d(o->SelfTransform, inv, o->GlobalTransform); }
    tnsExtractSelfTransformValue(o);
    tnsExtractGlobalTransformValue(o);
    if(ApplyToChild) for (laListItemPointer* li=o->ChildObjects.pFirst;li;li=li->pNext){ tnsSelfMatrixChanged(li->p,1); }
}
void tnsSelfTransformValueChanged(tnsObject* o){
    tnsMatrix44d Trans, Rot1, Rot2, Rot3, Scale, Res1, Res2;
    tnsLoadIdentity44d(o->SelfTransform);
    tnsMakeTranslationMatrix44d(Trans, LA_COLOR3(o->Location));
    tnsMakeScaleMatrix44d(Scale, o->Scale,o->Scale,o->Scale);

    tnsMakeRotationXMatrix44d(Rot1, o->Rotation[0]);
    tnsMakeRotationYMatrix44d(Rot2, o->Rotation[1]);
    tnsMakeRotationZMatrix44d(Rot3, o->Rotation[2]);
    switch (o->RotationMode){
    case TNS_ROTATION_ZYX_EULER: tnsMultiply44d(Res1, Trans, Rot1); tnsMultiply44d(Res2, Res1, Rot2); tnsMultiply44d(Res1, Res2, Rot3); break;
    case TNS_ROTATION_XZY_EULER: tnsMultiply44d(Res1, Trans, Rot1); tnsMultiply44d(Res2, Res1, Rot3); tnsMultiply44d(Res1, Res2, Rot2); break;
    case TNS_ROTATION_YXZ_EULER: tnsMultiply44d(Res1, Trans, Rot2); tnsMultiply44d(Res2, Res1, Rot1); tnsMultiply44d(Res1, Res2, Rot3); break;
    case TNS_ROTATION_YZX_EULER: tnsMultiply44d(Res1, Trans, Rot2); tnsMultiply44d(Res2, Res1, Rot3); tnsMultiply44d(Res1, Res2, Rot1); break;
    case TNS_ROTATION_ZXY_EULER: tnsMultiply44d(Res1, Trans, Rot3); tnsMultiply44d(Res2, Res1, Rot1); tnsMultiply44d(Res1, Res2, Rot2); break;
    case TNS_ROTATION_XYZ_EULER: tnsMultiply44d(Res1, Trans, Rot3); tnsMultiply44d(Res2, Res1, Rot2); tnsMultiply44d(Res1, Res2, Rot1); break;
    }
    tnsMultiply44d(o->SelfTransform, Res1, Scale);
    tnsSelfMatrixChanged(o,1);
}
void tnsGlobalTransformValueChanged(tnsObject* o){
    tnsMatrix44d Trans, Rot1, Rot2, Rot3, Scale, Res1, Res2;
    tnsLoadIdentity44d(o->GlobalTransform);
    tnsMakeTranslationMatrix44d(Trans, LA_COLOR3(o->GLocation));
    tnsMakeScaleMatrix44d(Scale, o->GScale,o->GScale,o->GScale);

    tnsMakeRotationXMatrix44d(Rot1, o->GRotation[0]);
    tnsMakeRotationYMatrix44d(Rot2, o->GRotation[1]);
    tnsMakeRotationZMatrix44d(Rot3, o->GRotation[2]);
    switch (o->RotationMode){
    case TNS_ROTATION_ZYX_EULER: tnsMultiply44d(Res1, Trans, Rot1); tnsMultiply44d(Res2, Res1, Rot2); tnsMultiply44d(Res1, Res2, Rot3); break;
    case TNS_ROTATION_XZY_EULER: tnsMultiply44d(Res1, Trans, Rot1); tnsMultiply44d(Res2, Res1, Rot3); tnsMultiply44d(Res1, Res2, Rot2); break;
    case TNS_ROTATION_YXZ_EULER: tnsMultiply44d(Res1, Trans, Rot2); tnsMultiply44d(Res2, Res1, Rot1); tnsMultiply44d(Res1, Res2, Rot3); break;
    case TNS_ROTATION_YZX_EULER: tnsMultiply44d(Res1, Trans, Rot2); tnsMultiply44d(Res2, Res1, Rot3); tnsMultiply44d(Res1, Res2, Rot1); break;
    case TNS_ROTATION_ZXY_EULER: tnsMultiply44d(Res1, Trans, Rot3); tnsMultiply44d(Res2, Res1, Rot1); tnsMultiply44d(Res1, Res2, Rot2); break;
    case TNS_ROTATION_XYZ_EULER: tnsMultiply44d(Res1, Trans, Rot3); tnsMultiply44d(Res2, Res1, Rot2); tnsMultiply44d(Res1, Res2, Rot1); break;
    }
    tnsMultiply44d(o->GlobalTransform, Res1, Scale);
    tnsGlobalMatrixChanged(o,1);
}

void tnsSetCurrentRoot(tnsObject *o){
    if(o->Type!=TNS_OBJECT_ROOT) return;
    T->World.ActiveRoot=o;
    // memAssignRef(s, &s->CurrentObject, o); we can't use this apparently because World is not hyper 2.
    // XXX: Needs fake mempool/user, or T->World prop redesign because T is runtime allocated.
}
void tnsInitObjectBase(tnsObject *o, tnsObject *under, char *Name, int Type,
                        real AtX, real AtY, real AtZ,
                        real RotX, real RotY, real RotZ, real RotW, u8bit RotationMode,
                        real Scale){
    if (!o) return;
    strSafeSet(&o->Name, Name);
    o->Type = Type;
    o->Location[0] = AtX;
    o->Location[1] = AtY;
    o->Location[2] = AtZ;
    o->Rotation[0] = RotX;
    o->Rotation[1] = RotY;
    o->Rotation[2] = RotZ;
    o->Rotation[3] = RotW;
    o->RotationMode = RotationMode;
    o->Scale = Scale;
    o->Show = 1;
    o->DrawMode = GL_LINE_LOOP;
    tnsSelfTransformValueChanged(o);
    lstAppendItem(&T->World.AllObjects, o);
    if (under){ lstAppendPointer(&under->ChildObjects, o);
        tnsObject* root=under->Type==TNS_OBJECT_ROOT?under:under->InRoot; memAssignRef(o, &o->InRoot, root); 
    }
}
int tnsCheckParentable(tnsObject* child, tnsObject* parent){
    if(child==parent||parent->ParentObject==child)return 0; tnsObject* o=parent; while(o=o->ParentObject){ if(o==parent) return 0; } return 1;
}
void tnsParentObject(tnsObject *child, tnsObject *parent, int KeepTransform){
    if (!child || !parent || child->ParentObject==parent || parent->ParentObject==child || parent==child) return;

    if (child->ParentObject) lstRemovePointer(&child->ParentObject->ChildObjects, child);
    else lstRemovePointer(&child->InRoot->ChildObjects, child);

    lstAppendPointer(&parent->ChildObjects, child);
    memAssignRef(child, &child->ParentObject, parent);

    if(KeepTransform) tnsGlobalMatrixChanged(child, 1);
    else tnsSelfMatrixChanged(child, 1);
}
void tnsUnparentObject(tnsObject *o, int KeepTransform){
    if (!o || !o->ParentObject) return;
    lstRemovePointer(&o->ParentObject->ChildObjects, o);
    lstAppendPointer(&o->InRoot->ChildObjects, o);
    memAssignRef(o, &o->ParentObject, 0);

    if(KeepTransform) tnsGlobalMatrixChanged(o,1);
    else tnsSelfMatrixChanged(o,1);
}
void tnsRotateObjectLocalValues(tnsObject *o, real x, real y, real z){
    o->Rotation[0] += x; o->Rotation[1] += y; o->Rotation[2] += z;
    tnsSelfTransformValueChanged(o);
}
void tnsMoveObjectLocal(tnsObject *o, real x, real y, real z){
    tnsMatrix44d mat, res1, res2, res3;
    tnsMakeTranslationMatrix44d(res2, x, y, z);
    tnsMultiply44d(res1, o->SelfTransform, res2);
    memcpy(o->SelfTransform, res1, sizeof(tnsMatrix44d));
    tnsSelfMatrixChanged(o,1);
}
void tnsMoveObjectGlobal(tnsObject *o, real x, real y, real z){
    tnsMatrix44d mat, res1, res2, res3;
    tnsMakeTranslationMatrix44d(res2, x, y, z);
    tnsMultiply44d(res1, res2, o->GlobalTransform); //if reverse then local
    memcpy(o->GlobalTransform, res1, sizeof(tnsMatrix44d));
    tnsGlobalMatrixChanged(o,1);
}
void tnsRotateObjectGlobal(tnsObject *o, real x, real y, real z, real angle, real cx,real cy,real cz){
    tnsMatrix44d tback,tfwd,rot,rot2,res1,res2;
    real xs, ys, zs;
    tnsMakeRotationMatrix44d(rot,angle,x,y,z);
    tnsMakeRotationMatrix44d(rot2,-angle,x,y,z);//why?

    tnsMakeTranslationMatrix44d(tfwd, o->GLocation[0],o->GLocation[1],o->GLocation[2]);
    tnsInverse44d(tback,tfwd);
    tnsMultiply44d(res1,o->GlobalTransform,tback);
    tnsMultiply44d(res2,rot,res1);
    tnsMultiply44d(o->GlobalTransform,res2,tfwd);

    tnsVector3d delta; delta[0]=o->GLocation[0]-cx; delta[1]=o->GLocation[1]-cy; delta[2]=o->GLocation[2]-cz;
    tnsVector3d c; c[0]=cx; c[1]=cy; c[2]=cz; tnsVector3d r;
    tnsApplyRotation43d(r,rot2,delta);
    tnsVectorPlus3d(&o->GlobalTransform[12],c,r);
    //memcpy(o->GlobalTransform, res1, sizeof(tnsMatrix44d));
    tnsGlobalMatrixChanged(o,1);
}
void tnsRotateObjectLocal(tnsObject *o, real x, real y, real z, real angle, real gcx,real gcy,real gcz){
    tnsMatrix44d tback,tfwd,rot,res1,res2;
    real xs, ys, zs;
    tnsVector3d gp,gpt; gp[0]=gcx; gp[1]=gcy; gp[2]=gcz;
    tnsApplyTransform43d(gpt,o->GlobalTransform,gp);
    tnsMakeTranslationMatrix44d(tfwd, gpt[0]-gcx, gpt[0]-gcy, gpt[0]-gcz);
    tnsInverse44d(tback,tfwd);
    tnsMakeRotationMatrix44d(rot,angle,x,y,z);
    tnsMultiply44d(res1,o->SelfTransform,tback);
    tnsMultiply44d(res2,res1,rot);
    tnsMultiply44d(o->SelfTransform,res2,tfwd);
    tnsSelfMatrixChanged(o,1);
}
void tnsScaleObject(tnsObject *o, real fac, real cx,real cy,real cz){
    tnsMatrix44d sca,res1,res2;
    tnsMakeScaleMatrix44d(sca,fac,fac,fac);
    tnsMultiply44d(res1,o->GlobalTransform,sca);
    tnsVector3d delta; delta[0]=res1[12]-cx; delta[1]=res1[13]-cy; delta[2]=res1[14]-cz;
    tnsVector3d c; c[0]=cx; c[1]=cy; c[2]=cz;
    tnsVectorMultiSelf3d(delta, fac);
    tnsVectorPlus3d(&res1[12],c,delta);
    memcpy(o->GlobalTransform, res1, sizeof(tnsMatrix44d));
    tnsGlobalMatrixChanged(o,1);
}
void tnsZoomViewingCamera(tnsCamera *c, real Ratio){
    if (c->FocusDistance < 0.1) return;
    tnsMoveObjectLocal(c, 0, 0, -c->FocusDistance * Ratio);
    c->FocusDistance *= (1 - Ratio);
}
void tnsRotateViewingCamera(tnsCamera *c, real x, real z){
    tnsMoveObjectLocal(c, 0, 0, -c->FocusDistance);
    tnsRotateObjectLocalValues(c, x, 0, z);
    tnsMoveObjectLocal(c, 0, 0, c->FocusDistance);
}
void tnsGetCameraMovingDeltas(tnsCamera *c, int ViewportW, int ViewportH, real x, real y, tnsVector4d p){
    memset(p, 0, sizeof(tnsVector4d));
    tnsVector4d tp = {0};
    tnsMatrix44d combine, projection, projinv, vpinv, vp;

    tnsLoadIdentity44d(projection);
    tnsMakePerspectiveMatrix44d(projection, c->FOV, (real)ViewportW / (real)ViewportH, c->ZMin, c->ZMax);
    tnsMakeViewportMatrix44d(vp, ViewportW, ViewportH, c->ZMax, c->ZMin);
    tnsInverse44d(projinv, projection);
    tnsInverse44d(vpinv, vp);

    p[2] = -c->FocusDistance;
    p[3] = 1;
    tnsApplyTransform44d(tp, projection, p);
    tnsVectorMultiSelf3d(tp, 1 / tp[3]);
    tnsApplyTransform43d(p, vp, tp);
    p[0] += x;
    p[1] += y;

    tnsApplyTransform43d(tp, vpinv, p);
    tnsVectorMultiSelf3d(tp, tp[3]);
    tnsApplyTransform44d(p, projinv, tp);
}
void tnsTranslateViewingCamera(tnsCamera *c, int ViewportW, int ViewportH, real x, real y){
    tnsVector4d p={0};
    tnsGetCameraMovingDeltas(c,ViewportW, ViewportH, x,y,p);
    tnsMoveObjectLocal(c, p[0], p[1], 0);
}

tnsObject *tnsCreateRootObject(char *name){
    tnsWorld *w = &T->World;
    tnsObject *o = memAcquireHyper(sizeof(tnsObject));

    o->Type==TNS_OBJECT_ROOT;
    strSafeSet(&o->Name, name);
    lstAppendItem(&w->RootObjects, o);

    w->ActiveRoot = o; // XXX: same here

    return o;
}
void tnsDestroyRootObject(tnsObject *root){
    tnsWorld *w = &T->World;
    tnsObject *o, *NextO;
    w->ActiveRoot = root->Item.pPrev ? root->Item.pPrev : root->Item.pNext ? root->Item.pNext : 0;

    lstRemoveItem(&w->RootObjects, root);
    while (lstPopPointer(&root->ChildObjects));
    strSafeDestroy(&root->Name);
    memFree(root);
}
void tnsDestroyObject(tnsObject *o){
    if(o->Type==TNS_OBJECT_ROOT){ tnsDestroyRootObject(o); return; }

    tnsUnparentObject(o,1);
    while(o->ChildObjects.pFirst){ tnsUnparentObject(((laListItemPointer*)o->ChildObjects.pFirst)->p, 1); }

    lstRemoveItem(&T->World.AllObjects, o);

    if(o->Type==TNS_OBJECT_MESH){ tnsInvaliateMeshBatch(o); }

    strSafeDestroy(&o->Name);
    memFree(o);
}
tnsCamera *tnsCreateCamera(tnsObject *under, char *Name, real FOV,
                        real AtX, real AtY, real AtZ,
                        real RotX, real RotY, real RotZ,
                        real FocusDistance){
    tnsCamera *c; tnsWorld *w = &T->World;

    c = memAcquireHyper(sizeof(tnsCamera));
    tnsInitObjectBase(&c->Base, under, Name, TNS_OBJECT_CAMERA, AtX, AtY, AtZ, RotX, RotY, RotZ, 1.0f, TNS_ROTATION_XYZ_EULER, 1.0f);
    c->FOV = FOV;
    c->CameraType = TNS_PRESPECTIVE_CAMERA;
    c->ZMin = 0.1f;
    c->ZMax = 1000.0f;
    c->FocusDistance = FocusDistance;
    c->OrthScale = 1.0f;

    return c;
}
tnsObject *tnsCreateEmpty(tnsObject *under, char *Name, real AtX, real AtY, real AtZ){
    tnsObject *o; tnsWorld *w = &T->World; if (!under) return 0;

    o = memAcquireHyper(sizeof(tnsObject));
    tnsInitObjectBase(o, under, Name, TNS_OBJECT_EMPTY, AtX, AtY, AtZ, 0, 0, 0, 1.0f, TNS_ROTATION_XYZ_EULER, 1.0f);

    return o;
}
tnsLight *tnsCreateLight(tnsObject *under, char *Name, real AtX, real AtY, real AtZ, real Strength, int UniDirectional){
    tnsLight *l; tnsWorld *w = &T->World; if (!under) return 0;
    tnsObject* root=under->Type==TNS_OBJECT_ROOT?under:under->InRoot;
    if (UniDirectional&&under){
        if(root){ if(root->ParentObject) return root->ParentObject; }
    }

    l = memAcquireHyper(sizeof(tnsLight));
    tnsInitObjectBase(&l->Base, under, Name, TNS_OBJECT_LIGHT, AtX, AtY, AtZ, 0, 0, 0, 1.0f, TNS_ROTATION_XYZ_EULER, 1.0f);
    l->Strength = Strength;
    l->UniDirectional = UniDirectional;

    if (root && UniDirectional && !root->ParentObject) memAssignRef(root, &root->ParentObject,l);
    return l;
}

int tnsAnyObjectsSelected(tnsObject *root){
    for(laListItemPointer* lip=root->ChildObjects.pFirst;lip;lip=lip->pNext){ tnsObject* o=lip->p; if(o->Flags&TNS_OBJECT_FLAGS_SELECTED) return 1; 
        if(o->ChildObjects.pFirst && tnsAnyObjectsSelected(o)) return 1;
    } return 0;
}
void tnsDeselectAllObjects(tnsObject *root){
    for(laListItemPointer* lip=root->ChildObjects.pFirst;lip;lip=lip->pNext){ tnsObject* o=lip->p; o->Flags&=(~TNS_OBJECT_FLAGS_SELECTED);
        if(o->ChildObjects.pFirst)tnsDeselectAllObjects(o);
    }
}
void tnsSelectAllObjects(tnsObject *root){
    for(laListItemPointer* lip=root->ChildObjects.pFirst;lip;lip=lip->pNext){ tnsObject* o=lip->p; o->Flags|=TNS_OBJECT_FLAGS_SELECTED;
        if(o->ChildObjects.pFirst)tnsSelectAllObjects(o);
    }
}
void tnsSelectObject(tnsObject* o, int Select, int Toggle){
    if(!o) return;
    if(Toggle) tnsSelectObject(o,(o->Flags&TNS_OBJECT_FLAGS_SELECTED?0:1),0);
    elif(Select) o->Flags|=TNS_OBJECT_FLAGS_SELECTED; else o->Flags&=(~TNS_OBJECT_FLAGS_SELECTED);
}

void tnsGetCameraViewProjection(tnsMatrix44d* mat, int w, int h, tnsCamera* Camera){
    tnsMatrix44d inv, result;
    tnsMakePerspectiveMatrix44d(mat, Camera->FOV, (real)w / (real)h, Camera->ZMin, Camera->ZMax);
    tnsInverse44d(inv, Camera->Base.GlobalTransform);
    tnsMultiply44d(result, mat, inv);
    memcpy(mat, result, sizeof(tnsMatrix44d));
}
void tnsApplyCameraView(int W, int H, tnsCamera *Camera){
    tnsMatrixStackItem *tmsi = tKnlGetCurrentMatStackItem();
    tnsShader *current_shader = 0;
    real *mat;
    tnsMatrix44d result, inv;

    if (!Camera) return;

    mat = tnsGetProjectionMatrix();
    tnsMakePerspectiveMatrix44d(mat, Camera->FOV, (real)W / (real)H, Camera->ZMin, Camera->ZMax);
    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyProjection(current_shader, mat);
        tnsShaderApplyProjectionInverse(current_shader, mat);
    }

    tnsResetViewMatrix();
    mat = tnsGetViewMatrix();
    tnsInverse44d(inv, Camera->Base.GlobalTransform);
    tnsMultiply44d(result, mat, inv);
    memcpy(mat, result, sizeof(tnsMatrix44d));

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyView(current_shader, result);
    }
}
void tnsApplyShadowCameraView(tnsLight *Light){
    tnsMatrixStackItem *tmsi = tKnlGetCurrentMatStackItem();
    tnsShader *current_shader = 0;
    tnsMatrix44d mat;
    tnsMatrix44d result, inv;

    if (!Light || !Light->UniDirectional) return;

    //mat = tnsGetProjectionMatrix();
    tnsMakeOrthoMatrix44d(mat, -1000,1000,-1000,1000,-1000,1000);

    tnsInverse44d(inv, Light->Base.GlobalTransform);
    tnsMultiply44d(result, mat, inv);
    memcpy(mat, result, sizeof(tnsMatrix44d));

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyShadowMatrix(current_shader,mat);
    }
}

void tnsApplyObjectMatrix(tnsObject *Object){
    tnsShader *current_shader = 0;
    real *mat;
    tnsMatrix44d result;

    if (!Object) return;

    mat = tnsGetModelMatrix();

    tnsMultiply44d(result, mat, Object->SelfTransform);
    memcpy(mat, result, sizeof(tnsMatrix44d));

    //Actually This Works Pretty Fast,but we are currently testing its functionality;
    //memcpy(mat, Object->GlobalTransform, sizeof(tnsMatrix44d));

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyModel(current_shader, mat);
    }
}

void tnsLookAt(tnsObject *o, tnsVector3d Target, tnsVector3d Up){
    tnsVector4d target = {0};
    tnsMatrix44d mat, res, res2;
    real *d;
    tnsShader *current_shader = 0;

    d = tnsGetModelMatrix();

    tnsMakeZTrackingMatrix44d(mat, o->GLocation, target, Up);
    tnsLoadIdentity44d(res);
    tnsMakeTranslationMatrix44d(res,o->GLocation[0],o->GLocation[1],o->GLocation[2]);
    tnsMultiply44d(res2, res, mat);
    memcpy(o->SelfTransform, res2, sizeof(tnsMatrix44d));

    tnsSelfMatrixChanged(o,1);

    if (current_shader = tKnlGetActiveShader()){
        tnsShaderApplyModel(current_shader, d);
    }
}

void tnsDrawCamera(tnsObject *o){
    tnsCamera *c = o;
    real fov_2 = c->FOV / 2;
    real ex, ey, ez;
    ey = 10 * sin(fov_2);
    ex = ey;
    ez = 10 * cos(fov_2);

    //if (T->CurrentShader != T->uiShader) tnsEnableShaderv(T->uiShader);

    tnsColor4d(1, 1, 1, 1);

    tnsVertex3d(ex, ey, -ez);
    tnsVertex3d(ex, -ey, -ez);
    tnsVertex3d(0.0, 0.0, 0.0);
    tnsVertex3d(ex, ey, -ez);
    tnsVertex3d(-ex, ey, -ez);
    tnsVertex3d(0.0, 0.0, 0.0);
    tnsVertex3d(-ex, -ey, -ez);
    tnsVertex3d(-ex, ey, -ez);

    tnsPackAs(GL_LINE_STRIP);
    tnsFlush();
}
void tnsDrawThisObject(tnsObject *o,tnsObject *active, int DrawAsObjectSelection, int MeshSelectionMode){
    if (!o->Show) return;
    switch (o->Type){
    case TNS_OBJECT_MESH:
        tnsDrawMeshObject(o, DrawAsObjectSelection, MeshSelectionMode, active);
        break;
    case TNS_OBJECT_CAMERA:
        tnsDrawCamera(o);
        break;
    case TNS_OBJECT_EMPTY:
    default:
        ////if (T->CurrentShader != T->uiShader) tnsEnableShaderv(T->uiShader);
        //tnsVertex3d(-1.0, 0.0, 0.0);
        //tnsVertex3d(5.0, 0.0, 0.0);
        //tnsVertex3d(0.0, 5.0, 0.0);
        //tnsVertex3d(0.0, -1.0, 0.0);
        //tnsVertex3d(0.0, 0.0, -1.0);
        //tnsVertex3d(0.0, 0.0, 5.0);
        //tnsColor4d(1, 1, 1, 1);
        //tnsPackAs(GL_LINES);
        //tnsFlush();
        break;
    }
}

void tnsDrawObjectTree(tnsObject *from, tnsObject *active, int DrawAsObjectSelection, int MeshSelectionMode){
    tnsObject *o; if(!from) return;

    for (laListItemPointer* lip=from->ChildObjects.pFirst;lip;lip=lip->pNext){
        o=lip->p;
        tnsPushMatrix(); tnsApplyObjectMatrix(o);
        tnsDrawThisObject(o, active, DrawAsObjectSelection, MeshSelectionMode);
        if (o->ChildObjects.pFirst){ tnsDrawObjectTree(o, active, DrawAsObjectSelection, MeshSelectionMode); }
        tnsPopMatrix();
    }
}

//===================================================================[Material]

tnsMaterial *tnsCreateMaterial(char *Name){
    tnsMaterial *m; if(!Name || !Name[0]) return 0;

    m = memAcquireHyper(sizeof(tnsMaterial));
    strSafeSet(&m->Name, Name);
    lstAppendItem(&T->World.Materials, m);
    return m;
}
tnsMaterial *tnsFindMaterialByIndex(int index){
    tnsMaterial *m; int i = 0;
    for (m = T->World.Materials.pFirst; m; m = m->Item.pNext){
        if (i == index){ m->ID = i; return m; } i++;
    }
    return 0;
}
tnsMaterial *tnsFindMaterial(char *name){
    tnsMaterial *m;
    int i = 0;
    for (m = T->World.Materials.pFirst; m; m = m->Item.pNext){
        if (strSame(m->Name->Ptr, name)){ return m; }
    }
    return 0;
}

//==================================================================[Util]

#define MAX3(a, b, c) \
    a > b ? (a > c ? a : c) : (b > c ? b : c)

#define MIN3(a, b, c) \
    a < b ? (a < c ? a : c) : (b < c ? b : c)

extern LA MAIN;

void tnsSingleLinearToLog(real *a, real gamma){
    if (*a < 0) *a = 0;
    *a = powf(*a, gamma);
}
void tnsSingleLogToLinear(real *a, real gamma){
    if (*a < 0) *a = 0;
    *a = powf(*a, 1.0f / gamma);
}
void tnsLinearToLog(real *rgb, real gamma){
    if (rgb[0] < 0) rgb[0] = 0;
    if (rgb[1] < 0) rgb[1] = 0;
    if (rgb[2] < 0) rgb[2] = 0;
    rgb[0] = powf(rgb[0], gamma);
    rgb[1] = powf(rgb[1], gamma);
    rgb[2] = powf(rgb[2], gamma);
}
void tnsLogToLinear(real *rgb, real gamma){
    if (rgb[0] < 0) rgb[0] = 0;
    if (rgb[1] < 0) rgb[1] = 0;
    if (rgb[2] < 0) rgb[2] = 0;
    rgb[0] = powf(rgb[0], 1.0f / gamma);
    rgb[1] = powf(rgb[1], 1.0f / gamma);
    rgb[2] = powf(rgb[2], 1.0f / gamma);
}
void tnsRgbToLuminance(real *rgb){
    real l;
    if (rgb[0] < 0) rgb[0] = 0;
    if (rgb[1] < 0) rgb[1] = 0;
    if (rgb[2] < 0) rgb[2] = 0;
    l = rgb[0] * 0.299 + rgb[1] * 0.587 + rgb[2] * 0.114;
    rgb[0] = rgb[1] = rgb[2] = l;
}

static const double LA_LUMA_WEIGHTS[]={ 0.2126, 0.7152, 0.0722 };
void tnsHCYtoRGB(real *hcy, real *rgb){
    double red, green, blue;
    double Y_peak = 0., H_insec, X, m;
    int H_sec;

    double hue = hcy[0];
    double chroma = hcy[1];
    double luma = hcy[2];
    double* weights=LA_LUMA_WEIGHTS;

    if (chroma < DBL_EPSILON){
        red = green = blue = luma;
    }else{
        hue = fmod(hue, 1.0);
        hue += hue < 0.0;
        hue *= 6.0;

        H_sec = (int)hue;

        switch (H_sec){
        case 0:
            H_insec = hue - H_sec;
            Y_peak = weights[0] + H_insec * weights[1];
            chroma *= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);
            X = chroma * H_insec;
            m = luma - (weights[0] * chroma + weights[1] * X);
            red = m + chroma;
            green = m + X;
            blue = m;
            break;
        case 1:
            H_insec = 1. - (hue - H_sec);
            Y_peak = weights[1] + H_insec * weights[0];
            chroma *= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);
            X = chroma * H_insec;
            m = luma - (weights[0] * X + weights[1] * chroma);
            red = m + X;
            green = m + chroma;
            blue = m;
            break;
        case 2:
            H_insec = hue - H_sec;
            Y_peak = weights[1] + H_insec * weights[2];
            chroma *= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);
            X = chroma * H_insec;
            m = luma - (weights[1] * chroma + weights[2] * X);
            red = m;
            green = m + chroma;
            blue = m + X;
            break;
        case 3:
            H_insec = 1. - (hue - H_sec);
            Y_peak = weights[2] + H_insec * weights[1];
            chroma *= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);
            X = chroma * H_insec;
            m = luma - (weights[1] * X + weights[2] * chroma);
            red = m;
            green = m + X;
            blue = m + chroma;
            break;
        case 4:
            H_insec = hue - H_sec;
            Y_peak = weights[2] + H_insec * weights[0];
            chroma *= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);
            X = chroma * H_insec;
            m = luma - (weights[0] * X + weights[2] * chroma);
            red = m + X;
            green = m;
            blue = m + chroma;
            break;
        default:
            H_insec = 1. - (hue - H_sec);
            Y_peak = weights[0] + H_insec * weights[2];
            chroma *= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);
            X = chroma * H_insec;
            m = luma - (weights[0] * chroma + weights[2] * X);
            red = m + chroma;
            green = m;
            blue = m + X;
            break;
        }
    }

    rgb[0] =red;
    rgb[1] =green;
    rgb[2] =blue;
}
void tnsRGBtoHCY(real *rgb, real *hcy){
    double hue, chroma, luma;
    double X, Y_peak = 0.;
    int H_sec = 4, t = -1;
    double* weights=LA_LUMA_WEIGHTS;

    int ix[3] = {0, 1, 2};

    if (rgb[0] < rgb[1]){
        if (rgb[1] > rgb[2]){
            if (rgb[0] < rgb[2]){
                ix[1] = 2;
                ix[2] = 1;
                H_sec = 2;
                t = 1;
            }else{
                ix[0] = 2;
                ix[1] = 0;
                ix[2] = 1;
                H_sec = 2;
            }
        }
    }else{
        if (rgb[1] < rgb[2]){
            if (rgb[0] < rgb[2]){
                ix[0] = 1;
                ix[1] = 0;
                H_sec = 4;
                t = 1;
            }else{
                ix[0] = 1;
                ix[1] = 2;
                ix[2] = 0;
                H_sec = 6;
            }
        }else{
            ix[0] = 2;
            ix[2] = 0;
            H_sec = 0;
            t = 1;
        }
    }

    luma = weights[0] * rgb[0] + weights[1] * rgb[1] + weights[2] * rgb[2];
    chroma = rgb[ix[2]] - rgb[ix[0]];

    if (chroma >= DBL_EPSILON){
        X = (rgb[ix[1]] - rgb[ix[0]]) / chroma;

        Y_peak = weights[ix[2]] + X * weights[ix[1]];
        if (luma != 0. && luma != 1.) chroma /= luma < Y_peak ? luma / Y_peak : (1. - luma) / (1. - Y_peak);

        hue = (H_sec + t * X) / 6.;
    }else
        chroma = hue = 0.0;

    hcy[0] = hue;
    hcy[1] = chroma;
    hcy[2] = luma;
}

void tnsClearAll(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}
void tnsClearColorv(real *rgba){
    glClearColor(rgba[0], rgba[1], rgba[2], rgba[3]);
}
void tnsClearColor(real r, real g, real b, real a){
    glClearColor(r, g, b, a);
}

//Assuming *result is long enough
void tnsMakeIndexUInt(unsigned int *result, int num, ...){
    int i;
    va_list list;

    va_start(list, num);

    for (i = 0; i < num; i++){
        result[i] = va_arg(list, unsigned int);
    }

    va_end(list);
}
void tnsMakeTriangle(real *arr, real x1, real y1, real x2, real y2, real x3, real y3){
    arr[0] = x1;
    arr[1] = y1;
    arr[2] = x2;
    arr[3] = y2;
    arr[4] = x3;
    arr[5] = y3;
}
void tnsMakeQuad2d(real *arr, real x1, real y1, real x2, real y2, real x3, real y3, real x4, real y4){
    arr[0] = x1;
    arr[1] = y1;
    arr[2] = x2;
    arr[3] = y2;
    arr[4] = x3;
    arr[5] = y3;
    arr[6] = x4;
    arr[7] = y4;
}
void tnsMakeQuadT2d(real *arr, real x1, real y1, real x2, real y2, real x3, real y3, real x4, real y4){
    arr[0] = x1;
    arr[1] = y1;
    arr[2] = x2;
    arr[3] = y2;

    arr[4] = x3;
    arr[5] = y3;
    arr[6] = x2;
    arr[7] = y2;
    
    arr[8] = x3;
    arr[9] = y3;
    arr[10] = x4;
    arr[11] = y4;
}
void tnsMakeQuad3d(real *arr, real x1, real y1, real z1, real x2, real y2, real z2, real x3, real y3, real z3, real x4, real y4, real z4){
    arr[0] = x1;
    arr[1] = y1;
    arr[2] = z1;
    arr[3] = x2;
    arr[4] = y2;
    arr[5] = z2;
    arr[6] = x3;
    arr[7] = y3;
    arr[8] = z3;
    arr[9] = x4;
    arr[10] = y4;
    arr[11] = z4;
}
void tnsMakeQuad4d(real *arr, real x1, real y1, real z1, real w1, real x2, real y2, real z2, real w2, real x3, real y3, real z3, real w3, real x4, real y4, real z4, real w4){
    arr[0] = x1;
    arr[1] = y1;
    arr[2] = z1;
    arr[3] = w1;
    arr[4] = x2;
    arr[5] = y2;
    arr[6] = z2;
    arr[7] = w2;
    arr[8] = x3;
    arr[9] = y3;
    arr[10] = z3;
    arr[11] = w3;
    arr[12] = x4;
    arr[13] = y4;
    arr[14] = z4;
    arr[15] = w4;
}
void tnsMakeCircle2d(real *arr, int slices, real ctrX, real ctrY, real r){
    int i;
    real radstep = 2 * TNS_PI / (real)slices;
    real rd = 0;
    real x, y;
    for (i = 0; i < slices; i++){
        x = ctrX + cos(rd) * r;
        y = ctrY + sin(rd) * r;
        arr[2 * i] = x;
        arr[2 * i + 1] = y;
        rd += radstep;
    }
}
void tnsMakeArc2d(real *arr, int slices, real ctrX, real ctrY, real r, real rad_begin, real rad_end){
    int i;
    real radstep = (rad_end - rad_begin) / (real)slices;
    real rd = rad_begin;
    real x, y;
    for (i = 0; i <= slices; i++){
        x = ctrX + cos(rd) * r;
        y = ctrY + sin(rd) * r;
        arr[2 * i] = x;
        arr[2 * i + 1] = y;
        rd += radstep;
    }
}

void tnsMakeBridgedIndex(unsigned int *result, int num, int revert, int begin){
    int i;
    if (!revert){
        for (i = 0; i < num; i++){
            result[i * 2] = begin + i;
            result[i * 2 + 1] = begin + i + num;
        }
    }else{
        for (i = 0; i < num; i++){
            result[i * 2] = begin + i;
            result[i * 2 + 1] = begin - i + num * 2;
        }
    }
}

void tnsMakeRing2d(real *arr, int *index, int slices, real ctrX, real ctrY, real r1, real r2){
    tnsMakeCircle2d(arr, slices, ctrX, ctrY, r1);
    tnsMakeCircle2d(&arr[slices*2], slices, ctrX, ctrY, r2);
    tnsMakeBridgedIndex(index, slices, 0, 0);
    index[slices*2]=0; index[slices*2+1]=slices;
}

void tnsMakeLinerGradient3d(real *arr, int num_points, real r0, real g0, real b0, real r1, real g1, real b1){
    int i = 0;
    real sr = (r1 - r0) / (real)num_points, sg = (g1 - g0) / (real)num_points, sb = (b1 - b0) / (real)num_points;
    for (i; i < num_points; i++){
        int a = 3 * i;
        arr[a] = r0 + sr * i;
        arr[a + 1] = g0 + sg * i;
        arr[a + 2] = b0 + sb * i;
    }
}
void tnsMakeLinerGradient4d(real *arr, int num_points, real r0, real g0, real b0, real a0, real r1, real g1, real b1, real a1){
    int i = 0;
    real sr = (r1 - r0) / (real)num_points, sg = (g1 - g0) / (real)num_points, sb = (b1 - b0) / (real)num_points, sa = (a1 - a0) / (real)num_points;
    for (i; i < num_points; i++){
        int a = 4 * i;
        arr[a] = r0 + sr * i;
        arr[a + 1] = g0 + sg * i;
        arr[a + 2] = b0 + sb * i;
        arr[a + 3] = a0 + sa * i;
    }
}
void tnsMakeLinerGradient3dv(real *arr, int num_points, real *rgb0, real *rgb1){
    int i = 0;
    real sr = (rgb1[0] - rgb0[0]) / (real)num_points, sg = (rgb1[1] - rgb0[1]) / (real)num_points, sb = (rgb1[2] - rgb0[2]) / (real)num_points;
    for (i; i < num_points; i++){
        int a = 3 * i;
        arr[a] = rgb0[0] + sr * i;
        arr[a + 1] = rgb0[1] + sg * i;
        arr[a + 2] = rgb0[2] + sb * i;
    }
}
void tnsMakeLinerGradient4dv(real *arr, int num_points, real *rgb0, real *rgb1){
    int i = 0;
    real sr = (rgb1[0] - rgb0[0]) / (real)(num_points - 1), sg = (rgb1[1] - rgb0[1]) / (real)(num_points - 1), sb = (rgb1[2] - rgb0[2]) / (real)(num_points - 1), sa = (rgb1[3] - rgb0[3]) / (real)(num_points - 1);
    for (i; i < num_points; i++){
        int a = 4 * i;
        arr[a] = rgb0[0] + sr * i;
        arr[a + 1] = rgb0[1] + sg * i;
        arr[a + 2] = rgb0[2] + sb * i;
        arr[a + 3] = rgb0[3] + sa * i;
    }
}

void tnsMakeFoucsSquare(int L, int R, int U, int B, int W){
    int w = W;
    //int al = Len + w;
    real v[16];

    //v[0] = L - w; v[1] = U - l;
    //v[2] = L - w; v[3] = U + w;
    //v[4] = L + l; v[4] = U + w;
    //v[6] = L + l; v[7] = U - w;
    //v[8] = L + w; v[9] = U - w;
    //v[10] = L + w; v[11] = U - l;

    //v[12] = R - l; v[13] = U + w;
    //v[14] = R + w; v[15] = U + w;
    //v[16] = R + w; v[17] = U - l;
    //v[18] = R - w; v[19] = U - l;
    //v[20] = R - w; v[21] = U - w;
    //v[22] = R - l; v[23] = U - w;

    tnsMakeQuad2d(v, L, U, R, U, R, B, L, B);
    tnsVertexArray2d(v, 4);
    tnsPackAs(GL_LINE_LOOP);
    tnsMakeQuad2d(&v[8], L + W, U - W, R - W, U - W, R - W, B + W, L + W, B + W);
    tnsVertexArray2d(&v[8], 4);
    tnsPackAs(GL_LINE_LOOP);
}

void tnsDrawFloor(int Size, int Span, int *ShowAxis){
    int i = 0;
    int Lim = Span * 2 + 1;
    int Dist = Size * Span;

    for (i; i < Lim; i++){
        if (i == Span && ShowAxis[0]) continue;
        tnsVertex3d(-Dist, i * Size - Dist, 0);
        tnsVertex3d(Dist, i * Size - Dist, 0);
    }

    for (i = 0; i < Lim; i++){
        if (i == Span && ShowAxis[1]) continue;
        tnsVertex3d(i * Size - Dist, -Dist, 0);
        tnsVertex3d(i * Size - Dist, Dist, 0);
    }

    tnsPackAs(GL_LINES);

    if (ShowAxis[0]){
        tnsColor4d(1, 0, 0, 1);
        tnsVertex3d(-Dist, 0, 0);
        tnsVertex3d(Dist, 0, 0);
        tnsPackAs(GL_LINES);
    }
    if (ShowAxis[1]){
        tnsColor4d(0, 1, 0, 1);
        tnsVertex3d(0, -Dist, 0);
        tnsVertex3d(0, Dist, 0);
        tnsPackAs(GL_LINES);
    }
    if (ShowAxis[2]){
        tnsColor4d(0, 0, 1, 1);
        tnsVertex3d(0, 0, -Dist);
        tnsVertex3d(0, 0, Dist);
        tnsPackAs(GL_LINES);
    }
}
